from glypy.composition.ccomposition cimport CComposition

from glypy._c.structure.glycan_composition cimport _CompositionBase

from glycopeptidepy._c.count_table cimport CountTable
from glycopeptidepy._c.structure.glycan cimport GlycanCompositionProxy, glycan_composition_type
from glycopeptidepy._c.structure.fragmentation_strategy.base cimport FragmentationStrategyBase
from glycopeptidepy._c.structure.fragment cimport _NameTree

from glycopeptidepy._c.structure.glycan cimport GlycosylationManager


cdef class GlycanCompositionFragment(object):
    cdef:
        public double mass
        public CComposition composition
        public CountTable key
        public bint is_extended
        public Py_hash_t _hash_key
        public object cache
        int _glycosylation_size

    cpdef GlycanCompositionFragment copy(self)
    cpdef _NameTree _get_glycan_composition(self)
    cdef int get_glycosylation_size(self)

    @staticmethod
    cdef GlycanCompositionFragment _create(double mass, CComposition composition, CountTable key, bint is_extended)


cdef class GlycanCompositionFragmentStrategyBase(FragmentationStrategyBase):
    cdef:
        public bint _use_query
        public object _generator
        public GlycosylationManager glycosylation_manager

    cpdef GlycanCompositionProxy glycan_composition(self)
    cdef bint _guess_query_mode(self, GlycanCompositionProxy glycan_composition)
    cpdef long count_glycosylation_type(self, glycotype)


cdef class StubGlycopeptideStrategy(GlycanCompositionFragmentStrategyBase):
    cdef:
        public bint extended
        public bint extended_fucosylation

    cpdef GlycanCompositionFragment fucosylate_increment(self, GlycanCompositionFragment shift)
    cpdef GlycanCompositionFragment xylosylate_increment(self, GlycanCompositionFragment shift)
    cpdef GlycanCompositionFragment modified_increment(self, GlycanCompositionFragment base, object shift)
    cpdef list fucosylate_extended(self, GlycanCompositionFragment shift, long fucose_count)
    cpdef bint _validate_glycan_composition(self, CountTable aggregate_glycosylation, glycan)
    cpdef list _combinate_sites(self, list per_site_shifts, object glycan)
    cdef list _combinate_sites_single(self, list per_site_shifts, object glycan)

    cpdef list n_glycan_composition_fragments(self, object glycan, int core_count=*, int iteration_count=*)
    cpdef list o_glycan_composition_fragments(self, glycan, long core_count=*, long iteration_count=*)
    cpdef list gag_linker_composition_fragments(self, glycan, long core_count=*, long iteration_count=*)

    cpdef list mixed_stub_fragments(self)
    cpdef list n_glycan_stub_fragments(self)
    cpdef list o_glycan_stub_fragments(self)
    cpdef list gag_linker_stub_fragments(self)

    cpdef list stub_fragments(self)


# cython: embedsignature=True

from collections import defaultdict
from itertools import product, combinations

cimport cython

from cpython.ref cimport Py_INCREF
from cpython cimport PyObject
from cpython.tuple cimport PyTuple_GetItem
from cpython.list cimport PyList_GetItem, PyList_SetItem, PyList_Size, PyList_New, PyList_GET_ITEM, PyList_GET_SIZE
from cpython.dict cimport (PyDict_GetItem, PyDict_SetItem, PyDict_Next,
                           PyDict_Keys, PyDict_Update, PyDict_DelItem, PyDict_Size)

from cpython.int cimport PyInt_AsLong, PyInt_Check, PyInt_FromLong

from cython.view cimport array as cvarray

from glypy.composition.ccomposition cimport CComposition

from glycopeptidepy._c.collectiontools cimport descending_combination_counter
from glycopeptidepy._c.count_table cimport CountTable, CountTableIterator
from glycopeptidepy._c.structure.constants cimport Configuration
from glycopeptidepy._c.structure.fragment cimport IonSeriesBase, PeptideFragment, ChemicalShiftBase
from glycopeptidepy._c.structure.sequence_methods cimport _PeptideSequenceCore
from glycopeptidepy._c.structure.base cimport ModificationBase, AminoAcidResidueBase, SequencePosition
from glycopeptidepy._c.structure.modification.modification cimport ModificationInstanceBase

from glycopeptidepy.structure import constants as _structure_constants
from glycopeptidepy.structure.modification import (
    ModificationCategory,
    ModificationIndex)

from glycopeptidepy._c.structure.fragmentation_strategy.base cimport (
    FragmentationStrategyBase,
    _n_glycosylation,
    _modification_hexnac,
    _o_glycosylation,
    _gag_linker_glycosylation,
    _modification_xylose,
    glycosylation_type_to_core)

from glycopeptidepy.utils.collectiontools import _AccumulatorBag

from glycopeptidepy.structure.modification import (
    Modification,
    NGlycanCoreGlycosylation,
    OGlycanCoreGlycosylation,
    GlycosaminoglycanLinkerGlycosylation)

from glycopeptidepy.structure.fragment import (
    ChemicalShift, IonSeries, format_negative_composition)

from glycopeptidepy.structure.glycan import (GlycosylationType)


cdef Configuration structure_constants = _structure_constants
cdef ModificationCategory_glycosylation = ModificationCategory.glycosylation

cdef ChemicalShiftBase ammonia_loss = ChemicalShift("-NH3", -CComposition({"N": 1, "H": 3}))
cdef ChemicalShiftBase water_loss = ChemicalShift("-H2O", -CComposition({"O": 1, "H": 2}))


cdef class PeptideFragmentationStrategyBase(FragmentationStrategyBase):

    def __init__(self, peptide, series, chemical_shifts=None, max_chemical_shifts=1, include_neutral_losses=False,
                 compute_compositions=False, **kwargs):
        if chemical_shifts is None:
            chemical_shifts = dict()
        elif not isinstance(chemical_shifts, dict):
            chemical_shifts = dict(chemical_shifts)
        super(PeptideFragmentationStrategyBase, self).__init__(peptide, compute_compositions, **kwargs)
        self.series = IonSeries(series)
        self.chemical_shift_rules = chemical_shifts
        self.max_chemical_shifts = max_chemical_shifts
        self.include_neutral_losses = include_neutral_losses
        self._initialize_fields()

    cpdef _initialize_fields(self):
        self.direction = self.series.direction

        # Null Values
        self.running_mass = 0
        self.running_delta_mass = 0
        if self.compute_compositions:
            self.running_composition = CComposition()
        else:
            self.running_composition = None
        self.index = -1
        self.size = self.peptide.get_size()

        self.modification_index = CountTable._create()
        self.glycosylation_manager = dict()
        self.amino_acids_counter = CountTable._create()

        self.running_mass += self.series.mass_shift
        if self.compute_compositions:
            self.running_composition.add_from(self.series.composition_shift)
        self._initialize_start_terminal()

    cpdef list _get_viable_chemical_shift_combinations(self):
        cdef:
            bint has_neutral_loss_rules
            list shifts
            PyObject* ptmp

        shifts = []
        if self.include_neutral_losses:
            shifts.append(ammonia_loss)
            shifts.append(water_loss)
        if not self.chemical_shift_rules:
            return shifts

        for residue, count in self.amino_acids_counter.items():
            ptmp = PyDict_GetItem(self.chemical_shift_rules, residue)
            if ptmp == NULL:
                pass
            else:
                loss_composition = <list?>ptmp
                if PyList_GET_SIZE(loss_composition):
                    for i in range(count):
                        shifts.append(loss_composition + [CComposition()])
        loss_composition_combinations = {}
        for comb in combinations(shifts, self.max_chemical_shifts):
            for prod in product(*comb):
                loss_composition = sum(prod, CComposition())
                loss_composition_combinations[format_negative_composition(loss_composition)] = loss_composition
        return [ChemicalShift(k, v) for k, v in loss_composition_combinations.items() if v]

    cpdef _initialize_start_terminal(self):
        if self.direction > 0:
            self.running_mass += self.peptide.get_n_term().mass
            if self.compute_compositions:
                self.running_composition.add_from(self.peptide.get_n_term().get_composition())
            self.index = -1
        elif self.direction < 0:
            self.running_mass += self.peptide.get_c_term().mass
            if self.compute_compositions:
                self.running_composition.add_from(self.peptide.get_c_term().get_composition())
            self.index = self.peptide.get_size()
        else:
            raise ValueError("Unknown direction %r" % (self.series.direction,))

    cpdef CComposition composition_of(self, SequencePosition position):
        cdef:
            size_t i, n
            CComposition composition
            ModificationBase mod

        n = SequencePosition.get_modification_count(position)
        composition = position.amino_acid.composition.clone()
        for i in range(n):
            mod = SequencePosition.get_modification(position.modifications, i)
            composition.add_from(mod.composition)
        return composition

    cpdef bint has_more(self):
        if self.direction > 0:
            return self.index < self.size - 2
        else:
            return self.index > 1

    cpdef list flanking_residues(self):
        cdef:
            list residues
            SequencePosition pos_a, pos_b

        residues = PyList_New(2)
        pos_a = self.peptide.get(self.index)
        pos_b = self.peptide.get(self.index + self.direction)
        if self.direction < 0:
            Py_INCREF(pos_b.amino_acid)
            PyList_SetItem(residues, 0, pos_b.amino_acid)
            Py_INCREF(pos_a.amino_acid)
            PyList_SetItem(residues, 1, pos_a.amino_acid)
        else:
            Py_INCREF(pos_a.amino_acid)
            PyList_SetItem(residues, 0, pos_a.amino_acid)
            Py_INCREF(pos_b.amino_acid)
            PyList_SetItem(residues, 1, pos_b.amino_acid)
        return residues

    cpdef long name_index_of(self):
        if self.direction > 0:
            return self.index + structure_constants.FRAG_OFFSET
        else:
            return (self.size - self.index - 1) + structure_constants.FRAG_OFFSET

    def __next__(self):
        if self.has_more():
            return self.step()
        else:
            raise StopIteration()

    cpdef track_glycosylation(self, long index, glycosylation):
        self.glycosylation_manager[self.index] = glycosylation
        self.modification_index.increment(glycosylation, 1)
        self.running_delta_mass += glycosylation.mass

    cpdef _update_state(self):
        cdef:
            SequencePosition position
            ModificationBase mod
            size_t i, n
        self.index += self.direction
        position = self.peptide[self.index]
        n = SequencePosition.get_modification_count(position)
        for i in range(n):
            mod =  SequencePosition.get_modification(position, i)
            if mod.is_tracked_for(ModificationCategory_glycosylation):
                self.track_glycosylation(self.index, mod)
            else:
                self.modification_index.increment(mod, 1)
                self.running_delta_mass += mod.mass
        # self.amino_acids_counter.increment(position.amino_acid, 1)
        self.running_mass += position.amino_acid.mass
        if self.compute_compositions:
            composition = self.composition_of(position)
            self.running_composition.add_from(composition)

    cpdef list _build_fragments(self):
        cdef:
            list fragments_from_site
            list partial_loss_fragments
            list shifts
            PeptideFragment fragment, f
            size_t i, n, j, m, k

        frag = PeptideFragment._create(
            self.series,
            self.name_index_of(),
            self.modification_index.copy(),
            self.running_mass,
            flanking_amino_acids=self.flanking_residues(),
            glycosylation=self.glycosylation_manager.copy() if self.glycosylation_manager else None,
            chemical_shift=None,
            composition=self.running_composition.clone() if self.compute_compositions else None,
            delta_mass=&self.running_delta_mass)

        shifts = None
        m = 0
        if PyDict_Size(self.chemical_shift_rules) or self.include_neutral_losses:
            shifts = self._get_viable_chemical_shift_combinations()
            m = PyList_GET_SIZE(shifts)

        partial_loss_fragments = self.partial_loss(frag)
        if not m:
            return partial_loss_fragments
        n = PyList_GET_SIZE(partial_loss_fragments)
        fragments_from_site = PyList_New(n * (m + 1))
        k = 0
        for i in range(n):
            fragment = <PeptideFragment>PyList_GET_ITEM(partial_loss_fragments, i)
            Py_INCREF(fragment)
            PyList_SetItem(fragments_from_site, k, fragment)
            k += 1
            for j in range(m):
                shift = <ChemicalShiftBase>PyList_GET_ITEM(shifts, j)
                f = fragment.clone()
                f.chemical_shift = shift
                Py_INCREF(f)
                PyList_SetItem(fragments_from_site, k, f)
                k += 1
        return fragments_from_site

    cpdef list partial_loss(self, PeptideFragment fragment):
        return [fragment]

    cpdef object step(self):
        self._update_state()
        fragments_from_site = self._build_fragments()
        return fragments_from_site

    def __repr__(self):
        return "%s(%s, %r, %r, %0.4f, %d)" % (
            self.__class__.__name__,
            self.peptide, self.series, self.running_composition,
            self.running_mass, self.index)

    cpdef reset(self):
        self._initialize_fields()


hcd_modifications_of_interest = {k.name: k for k in [
        _n_glycosylation,
        _modification_hexnac,
        _o_glycosylation,
        _gag_linker_glycosylation,
        _modification_xylose
    ]}

hcd_modification_compositions = {
    k.name: k.composition for k in hcd_modifications_of_interest.values()
}


hcd_modifications_of_interest_to_variants_cache = dict()


@cython.freelist(100)
cdef class ModificationConfiguration(object):

    @staticmethod
    cdef ModificationConfiguration _create(CountTable modifications_of_interest, CountTable other_modifications,
                                           CComposition delta_composition, CountTable modification_set,
                                           double other_modifications_mass):
        cdef ModificationConfiguration inst = ModificationConfiguration.__new__(ModificationConfiguration)
        inst.modifications_of_interest = modifications_of_interest
        inst.other_modifications = other_modifications
        inst.delta_composition = delta_composition
        inst.modification_set = modification_set
        inst.other_modifications_mass = other_modifications_mass
        return inst

    def __init__(self, modifications_of_interest, other_modifications, delta_composition, modification_set, other_modifications_mass):
        self.modifications_of_interest = modifications_of_interest
        self.other_modifications = other_modifications
        self.delta_composition = delta_composition
        self.modification_set = modification_set
        self.other_modifications_mass = other_modifications_mass

    def __iter__(self):
        yield self.modifications_of_interest
        yield self.delta_composition
        yield self.other_modifications

    def __repr__(self):
        return "{self.__class__.__name__}({self.modifications_of_interest}, {self.other_modifications})".format(
            self=self)

    def __eq__(self, other):
        if not isinstance(other, ModificationConfiguration):
            return NotImplemented
        else:
            return self.equal_to(<ModificationConfiguration>other)

    cdef bint equal_to(self, ModificationConfiguration other):
        if other is None:
            return False
        if not self.modifications_of_interest.equal_to(other.modifications_of_interest):
            return False
        if not self.other_modifications.equal_to(other.other_modifications):
            return False
        return True


cdef class HCDFragmentationStrategy(PeptideFragmentationStrategyBase):
    def __init__(self, peptide, series, chemical_shifts=None, max_chemical_shifts=1, include_neutral_losses=False, compute_compositions=False, **kwargs):
        super(HCDFragmentationStrategy, self).__init__(
            peptide, series, chemical_shifts, max_chemical_shifts, include_neutral_losses,
            compute_compositions, **kwargs)
        self._last_modification_set = None
        self._last_modification_variants = None

    cpdef _get_core_for(self, ModificationInstanceBase glycosylation):
        try:
            if not glycosylation.rule.is_core:
                glycosylation = glycosylation_type_to_core[glycosylation.rule.glycosylation_type]()
            return glycosylation
        except KeyError:
            raise ValueError("Cannot determine which core to use for {}".format(
                glycosylation.rule.glycosylation_type))

    @property
    def modification_variants_cache(self):
        return hcd_modifications_of_interest_to_variants_cache

    @modification_variants_cache.setter
    def modification_variants_cache(self, value):
        global hcd_modifications_of_interest_to_variants_cache
        hcd_modifications_of_interest_to_variants_cache = dict(value)

    cpdef CComposition composition_of(self, SequencePosition position):
        cdef:
            CComposition composition
            ModificationInstanceBase mod
            size_t i, n
        composition = position.amino_acid.composition.clone()
        n = SequencePosition.get_modification_count(position)
        for i in range(n):
            mod = <ModificationInstanceBase>SequencePosition.get_modification(position, i)
            if mod.is_tracked_for(ModificationCategory_glycosylation):
                mod = self._get_core_for(mod)
            composition.add_from(mod.composition)
        return composition

    cpdef track_glycosylation(self, long index, glycosylation):
        cdef:
            ModificationBase glycosylation_core

        # HCD strategy does not track intact topologies or compositions, just cores
        glycosylation_core = self._get_core_for(glycosylation)

        self.glycosylation_manager[self.index] = glycosylation_core
        self.modification_index.increment(glycosylation_core, 1)
        self.running_delta_mass += glycosylation_core.mass

    cpdef ModificationConfiguration _get_modifications_of_interest(self, PeptideFragment fragment):
        cdef:
            ModificationConfiguration result
            CComposition delta_composition
            CountTable modifications
            CountTableIterator modification_iterator
            CountTable modifications_of_interest, other_modifications
            Py_ssize_t pos, j
            PyObject *key
            ModificationBase mod
            long count
            double other_modifications_mass
        other_modifications_mass = 0
        if self._last_modification_set is not None:
            if self._last_modification_set.modification_set == fragment.modification_dict:
                return self._last_modification_set
        modifications = (fragment.modification_dict)
        if self.compute_compositions:
            delta_composition = CComposition._create(None)
        else:
            delta_composition = None
        other_modifications = CountTable._create()
        modifications_of_interest = CountTable._create()

        pos = 0
        modification_iterator = CountTableIterator._create(modifications)
        while modification_iterator.has_more():
            pos = modification_iterator.get_next_value(&key, &count)
            mod = <ModificationBase>key
            if mod.name in hcd_modifications_of_interest:
                modifications_of_interest.setitem(mod, count)
                if self.compute_compositions:
                    delta_composition.add_from(mod.composition * count)
            else:
                other_modifications.setitem(mod, count)
                other_modifications_mass += mod.mass * count
        result = ModificationConfiguration._create(
            modifications_of_interest, other_modifications,
            delta_composition, modifications, other_modifications_mass)
        return result

    cpdef _replace_cores(self, CountTable modifications_of_interest):
        cdef:
            long n_cores, o_cores, gag_cores, hexnac_cores
        n_cores = modifications_of_interest.delitem(_n_glycosylation)
        o_cores = modifications_of_interest.delitem(_o_glycosylation)
        gag_cores = modifications_of_interest.delitem(_gag_linker_glycosylation)

        # Convert core glycosylation into the residual monosaccharides remaining
        # after dissociation
        hexnac_cores = n_cores * 2 + o_cores
        if hexnac_cores:
            modifications_of_interest.increment(_modification_hexnac, hexnac_cores)
        if gag_cores:
            modifications_of_interest.increment(_modification_xylose, gag_cores)
        return modifications_of_interest

    cpdef list _generate_modification_variants(self, CountTable interesting_modifications, CountTable other_modifications):
        cdef:
            list variants, variant_modification_list
            CountTable varied_modifications
            CountTable updated_modifications
            CountTableIterator it
            size_t i, n, j
            Py_ssize_t pos
            PyObject *key
            PyObject *value
            ModificationBase mod
            double delta_mass
            long count
        variants = []
        variant_modification_list = descending_combination_counter(interesting_modifications)
        n = PyList_Size(variant_modification_list)
        for i in range(n):
            varied_modifications = <CountTable>PyList_GetItem(variant_modification_list, i)
            updated_modifications = CountTable._create_from(other_modifications)
            if self.compute_compositions:
                extra_composition = CComposition._create(None)
            else:
                extra_composition = None
            delta_mass = 0.0
            pos = 0
            j = 0
            it = CountTableIterator._create(varied_modifications)
            while it.has_more():
                pos = it.get_next_value(&key, &count)
                if pos != 0:
                    break
                j += 1
                mod = <ModificationBase>key
                if count != 0:
                    updated_modifications.increment(mod, count)
                    delta_mass += mod.mass * count
                    if self.compute_compositions:
                        extra_composition.add_from(mod.composition * count)
            variants.append((updated_modifications, extra_composition, delta_mass))
        return variants

    cpdef list partial_loss(self, PeptideFragment fragment):
        cdef:
            CComposition base_composition, delta_composition, new_composition
            ModificationConfiguration mod_config
            list variants, fragments
            IonSeriesBase series
            tuple variant_pair
            CountTable updated_modifications
            PyObject* ptemp
            double delta_mass
            size_t i, n
        mod_config = self._get_modifications_of_interest(fragment)
        if self.compute_compositions:
            base_composition = fragment.composition.clone()
            # remove the composition shift associated with the interesting modifications
            base_composition.subtract_from(mod_config.delta_composition)
        else:
            base_composition = None

        # If the modification configuration hasn't changed, reuse the existing variants.
        #
        # If this is the first time entering :meth:`partial_loss`, :attr:`_last_modification_set`
        # is :const:`None`, and :meth:`_get_modifications_of_interest` can never return :const:`None`
        # so this block will only be entered after at least one pass through the else clause.
        if mod_config is self._last_modification_set:
            variants = self._last_modification_variants
        else:
            # Recalculate the modifications according to the HCD rules, and update the instance's cache
            # properties.
            #
            # Check the class-wide variant cache to see if these variants have already been calculated,
            # and if so, reuse them, otherwise calculate them and update the class cache.
            self._replace_cores(mod_config.modifications_of_interest)
            self._last_modification_set = mod_config
            ptemp = PyDict_GetItem(
                hcd_modifications_of_interest_to_variants_cache,
                self._last_modification_set.modifications_of_interest)
            # if the combination is new, generate them and update the cache
            if ptemp == NULL:
                variants = self._last_modification_variants = self._generate_modification_variants(
                    mod_config.modifications_of_interest,
                    # do not include extra modifications here so that caching of
                    # interesting modificatoin variants does not need to be recalculated
                    # when uninteresting modifications change
                    CountTable._create()
                )
                PyDict_SetItem(
                    hcd_modifications_of_interest_to_variants_cache,
                    self._last_modification_set.modifications_of_interest,
                    variants)
            # otherwise reuse the cached version
            else:
                variants = self._last_modification_variants = <list>ptemp

        series = self.series

        n = PyList_Size(variants)
        # If there is only one variant, it must be the unaltered form, so just return that
        if n == 1:
            return [fragment]
        fragments = []
        for i in range(n):
            variant_pair = <tuple>PyList_GetItem(variants, i)
            updated_modifications = CountTable._create_from(<CountTable>PyTuple_GetItem(variant_pair, 0))
            # add the current fragment's uninteresting modifications back on top of the
            # varied modifications here.
            updated_modifications._add_from(mod_config.other_modifications)
            if self.compute_compositions:
                extra_composition = <CComposition>PyTuple_GetItem(variant_pair, 1)
                new_composition = base_composition.clone()
                new_composition.add_from(extra_composition)
            else:
                new_composition = None
            delta_mass = <object>PyTuple_GetItem(variant_pair, 2) + mod_config.other_modifications_mass
            fragments.append(
                PeptideFragment._create(
                    series, fragment.position, updated_modifications, fragment.bare_mass,
                    flanking_amino_acids=fragment.flanking_amino_acids,
                    glycosylation=None,
                    chemical_shift=None,
                    composition=new_composition,
                    delta_mass=&delta_mass
                ))
        return fragments

    @cython.boundscheck(False)
    cpdef double[::1] mass_series(self):
        cdef:
            double[::1] masses
            size_t i

        masses = cvarray(shape=(self.size - 1,), itemsize=sizeof(double), format='d')

        i = 0
        while self.has_more():
            self._update_state()
            masses[i] = self.running_mass
            i += 1
        return masses



import io

try:
    from urllib import urlopen
except ImportError:
    from urllib.request import urlopen

from . import fasta

peff_api_url = 'https://api.nextprot.org/export/entry/NX_%s.peff'


def download_peff(accession):
    url = peff_api_url % accession
    return fasta.PEFFReader(io.BytesIO(urlopen(url).read()), index=True)



import sys


def clean_definition(text):
    if text.startswith('"'):
        text = text.rsplit(" ", 1)[0]
        text = text[1:-1]
    return text


class Term(object):
    __slots__ = ("name", "id", "description", "category", "specialization")

    def __init__(self, name, id, description, category, specialization):
        self.name = name
        self.id = id
        self.description = description
        self.category = category
        self.specialization = specialization

    def __iter__(self):
        yield self.name
        yield self.id
        yield self.description
        yield self.category
        yield self.specialization

    def _asdict(self):
        return {
            "name": self.name,
            "id": self.id,
            "description": self.description,
            "category": self.category,
            "specialization": self.specialization
        }

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other or self.id == other
        else:
            return tuple(self) == tuple(other)

    def __str__(self):
        return str(self.name)

    def __repr__(self):
        text = "(%s)" % ', '.join("%s=%r" % (k, v) for k, v in self._asdict().items() if k != 'description')
        return self.__class__.__name__ + text

    def __reduce__(self):
        return self.__class__, (None, None, None, None, None), self.__getstate__()

    def __getstate__(self):
        return tuple(self)

    def __setstate__(self, d):
        if len(d) == 4:
            self.name, self.id, self.category, self.specialization = d
        else:
            self.name, self.id, self.description, self.category, self.specialization = d

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.name)

    def is_a(self, term):
        """Test whether this entity is exactly **term** or a specialization
        of **term**

        Parameters
        ----------
        term : str or :class:`~.Term`
            The entity to compare to

        Returns
        -------
        bool
        """
        return term == self.name or term in self.specialization


def _unique_list(items):  # pragma: no cover
    seen = set()
    out = []
    for x in items:
        if x in seen:
            continue
        seen.add(x)
        out.append(x)
    return out


class MappingProxy(object):
    def __init__(self, loader):
        assert callable(loader)
        self.loader = loader
        self.mapping = None

    def _ensure_mapping(self):
        if self.mapping is None:
            self.mapping = self.loader()

    def __getitem__(self, key):
        self._ensure_mapping()
        return self.mapping[key]


def _lazy_load_psims():
    try:
        from psims.controlled_vocabulary.controlled_vocabulary import load_psims
        cv_psims = load_psims()
    except Exception:  # pragma: no cover
        cv_psims = None
    return cv_psims


cv_psims = MappingProxy(_lazy_load_psims)


def type_path(term, seed):  # pragma: no cover
    path = []
    i = 0
    steps = []
    try:
        steps.append(term.is_a.comment)
    except AttributeError:
        steps.extend(t.comment for t in term.is_a)
    except KeyError:
        pass
    while i < len(steps):
        step = steps[i]
        i += 1
        path.append(step)
        term = cv_psims[step]
        try:
            steps.append(term.is_a.comment)
        except AttributeError:
            steps.extend(t.comment for t in term.is_a)
        except KeyError:
            continue
    return _unique_list(path)


def render_list(seed, list_name=None, term_cls_name="Term", writer=None):  # pragma: no cover
    if writer is None:
        writer = sys.stdout.write
    component_type_list = [seed]
    i = 0
    seen = set()
    if list_name is None:
        list_name = seed.replace(" ", "_") + 's'

    template = (
        "    %s(%r, %r, %r,"
        "       %r, %r), \n")
    writer("%s = [\n" % (list_name,))
    while i < len(component_type_list):
        component_type = component_type_list[i]
        i += 1
        for term in cv_psims[component_type].children:
            if term.name in seen:
                continue
            seen.add(term.name)
            writer(template % (
                term_cls_name, term.name, term.id, clean_definition(term.definition),
                component_type_list[0], type_path(term, seed)))
            if term.children:
                component_type_list.append(term.name)
    writer("]\n")


__all__ = [
    "Term", "cv_psims", "render_list",
    "MappingProxy"
]



from glypy.utils.cenum cimport EnumValue

from glycopeptidepy.structure.modification.descriptors import SequenceLocation, ModificationCategory


cdef EnumValue SequenceLocation_anywhere = SequenceLocation.anywhere


cdef class ModificationTargetBase(object):
    cdef:
        public frozenset amino_acid_targets
        public EnumValue position_modifier
        public list classification



cpdef tuple valid_site(ModificationTargetBase self, amino_acid=None, EnumValue position_modifier=SequenceLocation_anywhere):
    cdef:
        bint valid
    valid = False

    raise ValueError("Requires updating")

    # Validate amino acid target target
    valid = (self.amino_acid_targets is None) or (
        amino_acid in self.amino_acid_targets)
    valid = valid and ((self.position_modifier is SequenceLocation_anywhere) or
                        (position_modifier == self.position_modifier))

    # If the rule includes a position modifier other than anywhere
    if valid and (position_modifier is not SequenceLocation_anywhere) and (
            self.position_modifier is not SequenceLocation_anywhere):
        if position_modifier == self.position_modifier:
            if self.amino_acid_targets is None:
                return valid, position_modifier
            else:
                return valid, SequenceLocation_anywhere
        else:
            valid = False
            return valid, SequenceLocation_anywhere
    return valid, SequenceLocation_anywhere


from glypy.composition.ccomposition cimport CComposition
from glypy._c.structure.glycan_composition cimport _CompositionBase

from glycopeptidepy._c.structure.base cimport PeptideSequenceBase

ctypedef fused glycan_composition_type:
    _CompositionBase
    GlycanCompositionProxy
    object


cdef class GlycanCompositionProxy(object):
    '''A mapping-like object that imitates the GlycanComposition interface in
    a read-only fashion.
    '''

    cdef:
        public _CompositionBase obj
        public str _serialized

    cpdef object _getitem_fast(self, key)

    cpdef str serialize(self)

    @staticmethod
    cdef GlycanCompositionProxy _create(glycan_composition_type obj)


cdef class GlycanCompositionWithOffsetProxy(GlycanCompositionProxy):
    cdef:
        public CComposition composition_offset


cdef class GlycosylationManager(object):
    cdef:
        public dict mapping
        public object parent
        public object _aggregate
        public GlycanCompositionProxy _proxy
        public dict _type_track
        public int _total_glycosylation_size

    cpdef update(self, other)
    cpdef clear(self)
    cpdef keys(self)
    cpdef values(self)
    cpdef items(self)
    cpdef pop(self, key, default=*)

    cdef size_t get_size(self)

    cpdef GlycosylationManager copy(self)
    cpdef GlycosylationManager clone(self)

    cpdef object invalidate(self)

    cdef object get_aggregate(self)
    cdef int set_aggregate(self, object value)
    cpdef object _patch_aggregate(self)

    cdef dict get_glycosylation_types(self)
    cdef dict _update_track(self)

    cpdef CComposition total_composition(self)
    cpdef double mass(self)
    cpdef bint is_fully_specified_topologies(self)

    cdef GlycanCompositionProxy get_glycan_composition(self)
    cdef GlycanCompositionProxy _make_glycan_composition_proxy(self)
    cdef int get_total_glycosylation_size(self)
    cpdef int total_glycosylation_size(self)

    cpdef int count_glycosylation_type(self, glycosylation_type)

    cdef void _init(self)

    @staticmethod
    cdef GlycosylationManager _create(PeptideSequenceBase parent, aggregate)


cimport cython
from glypy.composition.ccomposition cimport CComposition

from glycopeptidepy._c.structure.fragment cimport IonSeriesBase
from glycopeptidepy._c.structure.sequence_methods cimport _PeptideSequenceCore
from glycopeptidepy._c.structure.base cimport ModificationBase


from glycopeptidepy.structure.modification import (
    Modification,
    NGlycanCoreGlycosylation,
    OGlycanCoreGlycosylation,
    GlycosaminoglycanLinkerGlycosylation)


from glycopeptidepy.structure.glycan import (GlycosylationType)


_n_glycosylation = NGlycanCoreGlycosylation()
_o_glycosylation = OGlycanCoreGlycosylation()
_gag_linker_glycosylation = GlycosaminoglycanLinkerGlycosylation()
_modification_hexnac = Modification("HexNAc").rule
_modification_xylose = Modification("Xyl").rule


glycosylation_type_to_core = {
    GlycosylationType.n_linked: _n_glycosylation,
    GlycosylationType.o_linked: _o_glycosylation,
    GlycosylationType.glycosaminoglycan: _gag_linker_glycosylation
}


@cython.freelist(1000)
cdef class FragmentationStrategyBase(object):
    def __init__(self, peptide, compute_compositions=False, *args, **kwargs):
        self.peptide = peptide
        self.compute_compositions = compute_compositions
        super(FragmentationStrategyBase, self).__init__(*args, **kwargs)

    def __repr__(self):
        return "%s(%s)" % (
            self.__class__.__name__,
            self.peptide)

    def next(self):
        return self.__next__()

    def __iter__(self):
        return self

    cpdef CComposition peptide_composition(self):
        return self.peptide.peptide_composition()

    cpdef CComposition total_composition(self):
        return self.peptide.total_composition()



'''Implementation of memoizing decorators
'''
from collections import OrderedDict
from functools import wraps


class MemoizedFunction(object):
    """A memoized function wrapper that manages its own hash table cache.

    The wrapped function's arguments must be hashable.

    Attributes
    ----------
    max_size: int
        The maximum number of distinct invocations to cache
    function: Callable
        The wrapped function
    cache: Mapping
        The hash table cache
    lru: bool
        Whether or not to enforce a LRU eviction strategy

    """
    def __init__(self, function, max_size=100, lru=False):
        self.max_size = max_size
        self.function = function
        self.lru = lru
        self._init_cache()
        wraps(function)(self)

    def _init_cache(self):
        if self.lru:
            self.cache = OrderedDict()
        else:
            self.cache = dict()

    def evict(self):
        if self.lru:
            self.cache.popitem(last=False)
        else:
            self.cache.popitem()

    def __call__(self, *args, **kwargs):
        key = (args, frozenset(kwargs.items()))
        if key not in self.cache:
            if len(self.cache) == self.max_size:
                self.evict()
            self.cache[key] = self.function(*args, **kwargs)
        if self.lru:
            value = self.cache.pop(key)
            self.cache[key] = value
            return value
        return self.cache[key]

    def clear(self):
        self.cache.clear()

    def __len__(self):
        return len(self.cache)

    def __repr__(self):
        template = "{self.__class__.__name__}({self.function}, {size}/{self.max_size})"
        size = len(self.cache)
        return template.format(self=self, size=size)

    @classmethod
    def memoize(cls, maxsize=100):
        """Make a memoization decorator. A negative value of `maxsize` means
        no size limit."""

        def _wrapper(func):
            return cls(func, maxsize)

        return _wrapper


memoize = MemoizedFunction.memoize


class FragmentCachingMixin(object):
    '''Implements a cache over get_fragments() that saves the last arguments and result
    value, and returns it instead of recomputing result if the next invocation's arguments
    match.
    '''
    def __init__(self, *args, **kwargs):
        self.__fragments_value = None
        self.__fragments_arguments = None
        super(FragmentCachingMixin, self).__init__(*args, **kwargs)

    def get_fragments(self, *args, **kwargs):
        if self.__fragments_arguments == (args, kwargs):
            return self.__fragments_value
        else:
            v = super(FragmentCachingMixin, self).get_fragments(*args, **kwargs)
            self.__fragments_value = tuple(v)
            self.__fragments_arguments = (args, kwargs)
            return self.__fragments_value

    def clear_fragments_cache(self):
        self.__fragments_value = None
        self.__fragments_arguments = None

    @classmethod
    def implement(cls, tp):
        return type(tp.__name__, (cls, tp), {})



from glycopeptidepy._c.structure.base cimport ModificationBase
from glycopeptidepy._c.structure.modification.rule cimport ModificationRuleBase


cdef class ModificationInstanceBase(ModificationBase):

    cpdef bint is_a(self, object category):
        return self.rule.is_a(category)

    cpdef basestring serialize(self):
        cdef:
            basestring rep
        if self.rule.is_standard:
            rep = self.name
        else:
            rep = self.rule.serialize()
        return rep

    cpdef _init_from_rule(self, ModificationRuleBase rule):
        self.name = rule.name
        self.mass = rule.mass
        self.rule = rule
        self._hash = self.rule._hash

        try:
            self.composition = rule.composition
        except AttributeError:
            self.composition = None

    cpdef bint is_tracked_for(self, category):
        """Determine if :attr:`rule` is tracked by a particular
        behavioral pattern associated with a :class:`~.ModificationCategory`.

        This relationship is distinct from :meth:`is_a` which merely observes that
        the semantic relationship holds, not that any actual behavior is available.

        Parameters
        ----------
        category : :class:`~.ModificationCategory`
            The category to check

        Returns
        -------
        bool
        """
        return self.rule.is_tracked_for(category)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if isinstance(self, ModificationInstanceBase):
            if isinstance(other, ModificationBase):
                return (<ModificationBase>other).name in (<ModificationInstanceBase>self).rule.names
            else:
                return other in (<ModificationInstanceBase>self).rule.names
        else:
            if isinstance(self, ModificationBase):
                return (<ModificationBase>self).name in (<ModificationInstanceBase>other).rule.names
            else:
                return self in (<ModificationInstanceBase>other).rule.names


    def __ne__(self, other):
        return not self == other


from glycopeptidepy._c.structure.base cimport ModificationBase
from glycopeptidepy._c.structure.modification.rule cimport ModificationRuleBase


cdef class ModificationInstanceBase(ModificationBase):
    cdef:
        public ModificationRuleBase rule
        public Py_hash_t _hash

    cpdef _init_from_rule(self, ModificationRuleBase rule)

    cpdef bint is_tracked_for(self, category)



from collections import Iterable

sentinel = object()


class peekable(Iterable):
    sentinel = sentinel

    def __init__(self, iterable):
        self.iterable = iter(iterable)
        try:
            self._lookahead = next(self.iterable)
        except StopIteration:
            self._lookahead = sentinel

    def __iter__(self):
        # Counter to determine if we've passed through the
        # inner loop at all
        i = 0
        # Default stopping value for x to make accidental reuse
        # of the iterator safer
        x = sentinel
        for x in self.iterable:
            val = self._lookahead
            self._lookahead = x
            yield (val, x)
            i += 1
            x = sentinel
        else:
            val = self._lookahead
            self._lookahead = x
        if i > 0:
            yield val, x
        return

    @property
    def peek(self):
        return self._lookahead



# pragma: no cover

import itertools

from six import string_types as basestring

from glypy import MonosaccharideResidue

from glycopeptidepy.utils.memoize import memoize

from .sequence import PeptideSequenceBase
from .residue import Residue as AminoAcidResidue, get_all_residues
from .composition import Composition
from .modification import Modification
from .base import SequencePosition

class AminoAcidSequenceBuildingBlock(object):
    @classmethod
    def get_all_common_residues(cls):
        return map(cls, get_all_residues())

    @classmethod
    def get_all_sequencing_residues(cls):
        residues = get_all_residues()
        for residue in list(residues):
            degenerate = residue.is_degenerate
            if degenerate:
                for degen in degenerate:
                    residues.remove(degen)
        return map(cls, residues)

    def __init__(self, residue_, modifications=None, mass=None):
        if modifications is None:
            modifications = ()
        self.residue = residue_
        self.modifications = tuple(modifications)
        if mass is None:
            mass = residue_.mass + sum(m.mass for m in modifications)
        self.mass = mass
        self._string = None

    def __iter__(self):
        yield self.residue
        yield self.modifications

    def _reset(self):
        self._string = None

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        try:
            return self.residue is other.residue and self.modifications == other.modifications
        except AttributeError:
            return other == str(self)

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return "({}, {}):{:.2f}".format(self.residue.symbol, self.modifications, self.mass)

    def __str__(self):
        if self._string is None:
            self._string = "{}{}".format(
                self.residue.symbol,
                "({})".format(self.modifications[0]) if len(self.modifications) > 0 else "")
        return self._string

    def __contains__(self, item):
        return self == item

    @classmethod
    @memoize()
    def from_str(cls, string):
        # A hack to get ModificationBuildingBlock into the same flow of execution
        # Move this logic to a pure function later.
        if string.startswith("@"):
            return ModificationBuildingBlock(string)
        parts = string.split("(")
        aa = AminoAcidResidue(parts[0])
        mod = tuple()
        if len(parts) > 1:
            mod = (Modification(parts[1][:-1]),)
        return cls(aa, mod)

    def orderings(self):
        yield (self)

    def clone(self):
        return self.__class__(self.residue, tuple(self.modifications))

    def as_sequence_position(self):
        return SequencePosition([self.residue, list(self.modifications)])

    @classmethod
    def from_sequence_position(cls, position):
        return cls(position.amino_acid, tuple(position.modifications))


class ModificationBuildingBlock(object):
    # Not all SequenceComposition methods are compatible with these?
    sigil = '@'

    def __init__(self, name):
        try:
            if name.startswith(self.sigil):
                self.name = name
                self.modification = Modification(name[1:])
            else:
                self.name = self.sigil + name
                self.modification = Modification(name)
        except AttributeError:
            self.name = self.sigil + str(name)
            self.modification = name
        self.mass = self.modification.mass
        self._hash = hash(self.modification)

    def __repr__(self):
        return self.name

    def __eq__(self, other):
        try:
            return self.modification == other.modification
        except AttributeError:
            return str(self) == str(other)

    def __ne__(self, other):
        try:
            return self.modification != other.modification
        except AttributeError:
            return str(self) != str(other)

    def __hash__(self):
        return self._hash

    def clone(self):
        return self.__class__(self.name)


n_hexnac = AminoAcidSequenceBuildingBlock.from_str("N(HexNAc)")


class MonosaccharideResidueAdapter(object):
    def __init__(self, residue):
        if isinstance(residue, basestring):
            residue = MonosaccharideResidue.from_iupac_lite(residue)
        self.residue = residue
        self.mass = residue.mass()
        self.symbol = "(%s)" % str(residue)

    def __hash__(self):
        return hash(self.residue)

    def __eq__(self, other):
        if isinstance(other, (MonosaccharideResidue, AminoAcidResidue)):
            return self.symbol == other.symbol
        else:
            return self.symbol == other

    def __ne__(self, other):
        return not self == other


class SequenceComposition(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self)
        self._mass = None
        self._composition_offset = Composition("H2O")
        self.extend(*args)

    def clone(self):
        inst = self.__class__()
        inst += self
        return inst

    def __setitem__(self, key, value):
        if key is None:
            return
        if isinstance(key, basestring):
            key = AminoAcidSequenceBuildingBlock.from_str(key)
        dict.__setitem__(self, key, value)
        self._mass = None

    def __getitem__(self, key):
        if isinstance(key, basestring):
            key = AminoAcidSequenceBuildingBlock.from_str(key)
        return dict.__getitem__(self, key)

    def __contains__(self, key):
        if isinstance(key, dict):
            for k, v in key.items():
                if self[k] - v < 0:
                    return False
            return True
        else:
            if isinstance(key, basestring):
                key = AminoAcidSequenceBuildingBlock.from_str(key)
            return dict.__contains__(self, key)

    @property
    def mass(self):
        if self._mass is not None:
            return self._mass
        mass = self._composition_offset.mass
        for residue_type, count in self.items():
            mass += residue_type.mass * count
        self._mass = mass
        return mass

    def extend(self, *args):
        for residue in args:
            if isinstance(residue, PeptideSequenceBase):
                for pos in residue:
                    self[AminoAcidSequenceBuildingBlock.from_sequence_position(pos)] += 1
            else:
                self[residue] += 1

    def __iadd__(self, other):
        for elem, cnt in (other.items()):
            self[elem] += cnt
        return self

    def __add__(self, other):
        result = self.clone()
        for elem, cnt in other.items():
            result[elem] += cnt
        return result

    def __radd__(self, other):
        return self + other

    def __isub__(self, other):
        for elem, cnt in other.items():
            self[elem] -= cnt
        return self

    def __sub__(self, other):
        result = self.clone()
        for elem, cnt in other.items():
            result[elem] -= cnt
        return result

    def __rsub__(self, other):
        return (self - other) * (-1)

    def __mul__(self, other):
        if not isinstance(other, int):
            raise TypeError(
                'Cannot multiply Composition by non-integer',
                other)
        prod = SequenceComposition()
        for k, v in self.items():
            prod[k] = v * other

        return prod

    def __rmul__(self, other):
        return self * other

    def __eq__(self, other):
        if not isinstance(other, dict):
            return False
        self_items = set([i for i in self.items() if i[1]])
        other_items = set([i for i in other.items() if i[1]])
        return self_items == other_items

    def __neg__(self):
        return -1 * self

    def __missing__(self, key):
        return 0

    @property
    def composition_offset(self):
        return self._composition_offset

    @composition_offset.setter
    def composition_offset(self, value):
        self._mass = None
        self._composition_offset = value

    def flatten(self):
        for k in self:
            for i in range(0, self[k]):
                yield k

    def serialize(self):
        return "{%s}" % '; '.join("{}:{}".format(str(k), v) for k, v in sorted(
            self.items(), key=lambda x: x[0].mass))

    __str__ = serialize

    def __hash__(self):
        return hash(str(self))

    @classmethod
    def parse(cls, string):
        inst = cls()
        tokens = string[1:-1].split('; ')
        for token in tokens:
            residue, count = token.split(":")
            inst[AminoAcidSequenceBuildingBlock.from_str(residue)] = int(count)
        return inst

    def maybe_n_glycosylation(self):
        n_N = self["N"] + self[n_hexnac]
        n_S = self["S"]
        n_T = self["T"]
        used = 0
        possible_sequons = 0
        while(n_N > 0):
            if n_S > 0 or n_T > 0:
                if sum(self.values()) - used >= 3:
                    n_N -= 1
                    if n_S > 0:
                        n_S -= 1
                    elif n_T > 0:
                        n_T -= 1
                    else:
                        break
                    used += 3
                    possible_sequons += 1
                    continue
            break
        return possible_sequons

    def orderings(self):
        for ordering in itertools.permutations(self.flatten()):
            yield ordering


def all_compositions(blocks, size=20):
    components = [None] + list(blocks)
    for case in itertools.combinations_with_replacement(components, size):
        sc = SequenceComposition(*case)
        if len(sc) == 0:
            continue
        yield sc



'''Represent rules describing peptide modifications with
particular chemical properties, specificities, and other
metadata.

The core data structure is :class:`ModificationRule`, which
in addition to its chemical composition, mass, target specificities,
and categorical descriptors, carries a substantial amount of information
about what its names are.
'''

import re

from collections import defaultdict
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

from six import string_types as basestring

from ..base import ModificationBase
from ..composition import Composition, formula
from ..residue import AminoAcidResidue

from .utils import ModificationNameResolutionError
from .descriptors import ModificationCategory, SequenceLocation
from .target import (
    extract_targets_from_string, ModificationTarget,
    title_cleaner, get_position_modifier_rules_dict)


class _ModificationResolver(object):
    def __init__(self, rule_types=None):
        if rule_types is None:
            rule_types = []
        self.rule_types = list(rule_types) or []

    def register(self, rule_type):
        """Register a :class:`~.ModificationRule` subclass with
        the parsing cascade this object represents.

        This method may be used as a decorator for target classes.

        Parameters
        ----------
        rule_type: :class:`ModificationRule` subclass
            The subclass to register

        Returns
        -------
        :class:`ModificationRule` subclass
            The `rule_type` value, allowing this method to be used
            as a decorator.
        """
        self.rule_types.append(rule_type)
        return rule_type

    def __call__(self, name):
        for rule_type in self.rule_types:
            result = rule_type.try_parse(name)
            if result is not None:
                return result
        raise ModificationNameResolutionError(name)


try:
    from glycopeptidepy._c.structure.modification.rule import ModificationRuleBase
except ImportError:
    class ModificationRuleBase(ModificationBase):
        """Represent common properties of modification rules.
        """
        def is_a(self, category):
            '''Returns whether or not this :class:`ModificationRule` object belongs to
            the specified :class:`~.ModificationCategory`.

            Returns
            -------
            bool
            '''
            return category in self.categories

        def __hash__(self):
            return self._hash

        def __eq__(self, other):
            if self is other:
                return True
            idents = self.names
            try:
                other_idents = other.names
            except AttributeError:
                other_idents = {other}
            return len(idents & other_idents) > 0

        def __ne__(self, other):
            return not self == other

        def is_tracked_for(self, category):
            """Determine if this :class:`ModificationRule` is tracked by a particular
            behavioral pattern associated with a :class:`~.ModificationCategory`.

            This relationship is distinct from :meth:`is_a` which merely observes that
            the semantic relationship holds, not that any actual behavior is available.

            Parameters
            ----------
            category : :class:`~.ModificationCategory`
                The category to check

            Returns
            -------
            bool
            """
            return False

        @property
        def is_standard(self):
            """Indicates whether the :class:`ModificationRule` describes a modification which
            has been specified from a reference source, and may be fully reconstructed from
            its name alone.

            This property is not strictly enforced, and does not cover all aspects of the object's
            state. It does however provide a general guide for whether or not a modification is able
            to be translated from :mod:`glycopeptidepy`'s internal representation into something that
            another program *should* be able to understand.

            Returns
            -------
            bool
            """
            return True

        def serialize(self):
            '''A string representation for inclusion in sequences'''
            return self.name


def _ModificationRule_reconstructor(tp):
    return tp.__new__(tp)


def _merge_dictlists(adict, bdict):
    result = dict()
    for key in set(adict) | set(bdict):
        result[key] = list(set(adict.get(key, [])) | set(bdict.get(key, [])))
    return result


class ModificationLocatingMixin(object):
    __slots__ = ()

    def clear_targets(self):
        self._n_term_targets = None
        self._c_term_targets = None
        self.targets = set()

    def _configure_targets(self, amino_acid_specificity):
        self._n_term_targets = None
        self._c_term_targets = None

        # The type of the parameter passed for amino_acid_specificity is variable
        # so select the method correct for the passed type

        # If the specificity is a string, parse it into rules
        if isinstance(amino_acid_specificity, basestring):
            try:
                self.targets = set([
                    extract_targets_from_string(amino_acid_specificity)])
            except TypeError:
                raise TypeError("Could not extract targets from %s" %
                                amino_acid_specificity)

        # If the specificity is already a rule, store it as a list of one rules
        elif isinstance(amino_acid_specificity, ModificationTarget):
            self.targets = set([amino_acid_specificity])

        # If the specificity is a collection of rules, store it as is
        elif isinstance(amino_acid_specificity, Iterable):
            self.targets = set(amino_acid_specificity)
        else:
            raise ValueError("Could not interpret target specificity from %r" % (
                amino_acid_specificity,))

    def valid_site(self, amino_acid=None, position_modifiers=None):
        valid_sites = []
        for target in self.targets:
            result, site = target.valid_site(amino_acid, position_modifiers)
            if not result:
                continue
            # if SequenceLocation.anywhere is encountered, store 0 since None
            # can't be ordered with numbers
            if site.value is None:
                valid_sites.append(0)
            else:
                # otherwise store the location's numeric value
                valid_sites.append(site.value)
        if not valid_sites:
            return False
        # Find the minimum site type. Terminal sites are negative numbers
        # and prefer those from 0 which denotes SequenceLocation.anywhere
        site_type = min(valid_sites)
        if site_type == 0:
            return SequenceLocation.anywhere
        return SequenceLocation[site_type]

    def why_valid(self, amino_acid=None, position_modifiers=None):
        possible_targets = [target for target in self.targets if target.valid_site(
            amino_acid, position_modifiers)[0]]
        if not possible_targets:
            return None
        minimized_target = min(possible_targets, key=len)
        return minimized_target

    def find_valid_sites(self, sequence, protein_n_term=False, protein_c_term=False):
        valid_indices = []
        position_modifier_rules = get_position_modifier_rules_dict(
            sequence, protein_n_term=protein_n_term, protein_c_term=protein_c_term)
        for index in range(len(sequence)):
            position = sequence[index]
            if position.modifications:
                continue
            amino_acid = position[0].name
            position_modifiers = position_modifier_rules.get(index)
            is_valid = self.valid_site(amino_acid, position_modifiers)
            if is_valid:
                if (is_valid is SequenceLocation.n_term) or (is_valid is SequenceLocation.c_term):
                    valid_indices.append(is_valid)
                else:
                    valid_indices.append(index)

        return valid_indices

    @property
    def n_term_targets(self):
        if self._n_term_targets is None:
            solutions = []
            for target in self.targets:
                if target.position_modifier == SequenceLocation.n_term or target.position_modifier == SequenceLocation.protein_n_term:
                    solutions.append(target)
            self._n_term_targets = solutions
        return self._n_term_targets

    @property
    def c_term_targets(self):
        if self._c_term_targets is None:
            solutions = []
            for target in self.targets:
                if target.position_modifier == SequenceLocation.c_term or target.position_modifier == SequenceLocation.protein_c_term:
                    solutions.append(target)
            self._c_term_targets = solutions
        return self._c_term_targets


class ModificationRule(ModificationRuleBase, ModificationLocatingMixin):
    '''Represent the name, mass, and position specifities associated with a given modification.
    Additionally stores information on metadata about this rule computed from these values and
    categorical information from expert sources.

    Attributes
    ----------
    categories : list
        The list of named categories of PTMs that this modification
        falls under. Examples are "artefact", "glycosylation", or "biological process"
    mass : float
        The mass of the modification defined by this rule
    name : str
        A common name for the modification defined by this rule
    names : set
        The set of all names used to refer to this modification, beyond those
        commonly used and those for display purposes.
    aliases : set
        The set of all names that are allowed to be used to display this modification
    options : dict
        Arbitrary information from external sources about this modification
        rule and its phenomena.
    title : str
        A formal or "full" name for this modification that may be its chemical name
        or some other more verbose representation which is used to identify it but
        not convenient to read or write.
    targets : set
        All applicable :class:`.ModificationTarget` instances which apply to this
        rule. This set may differ from instance to instance of the same rule by
        configuration as per the :class:`RestrictedModificationTable`, for example
    '''

    resolve = _ModificationResolver()

    @staticmethod
    def modification_tp(*args, **kwargs):
        from .modification import Modification
        return Modification(*args, **kwargs)

    @staticmethod
    def reduce(rules):
        """Given a collection of :class:`ModificationRule` instances, combine
        those with the same name into a single object using :meth:`ModificationRule.__add__`

        Parameters
        ----------
        rules : :class:`list` of :class:`ModificationRule`
            The rules to collapse

        Returns
        -------
        list
        """
        collector = defaultdict(list)
        for rule in rules:
            collector[rule.name].append(rule)
        reductions = []
        for _, rules in collector.items():
            accum = rules[0]
            for other in rules[1:]:
                accum = accum + other
            reductions.append(accum)
        return reductions

    @classmethod
    def from_protein_prospector(cls, amino_acid_specificity, modification_name,
                                title=None, monoisotopic_mass=None,
                                neutral_losses=None):
        """Build a :class:`ModificationRule` from data scraped from Protein Prospector

        Parameters
        ----------
        amino_acid_specificity : str
            The amino acid and position target specificity
        modification_name : str
            The name of the modification
        title : str, optional
            The display title for the modification, by default None
        monoisotopic_mass : float, optional
            The monoisotopic mass shift of the modification, by default None
        neutral_losses : list, optional
            Any neutral losses associated with the modification, by default None

        Returns
        -------
        :class:`ModificationRule`
        """
        if neutral_losses is None:
            neutral_losses = []
        # If the specificity is a string, parse it into rules
        targets = None
        if isinstance(amino_acid_specificity, basestring):
            try:
                targets = set(
                    [extract_targets_from_string(amino_acid_specificity)])
            except TypeError:
                raise TypeError("Could not extract targets for %s with specificity %r" % (
                    modification_name, amino_acid_specificity))

        # If the specificity is already a rule, store it as a list of one rules
        elif isinstance(amino_acid_specificity, ModificationTarget):
            targets = set([amino_acid_specificity])

        # If the specificity is a collection of rules, store it as is
        elif isinstance(amino_acid_specificity, Iterable) and\
                isinstance(iter(amino_acid_specificity).next(), ModificationTarget):
            targets = set(amino_acid_specificity)

        return cls(targets, modification_name, title, monoisotopic_mass,
                   neutral_losses=neutral_losses)

    @classmethod
    def from_unimod(cls, unimod_entry):
        """Createa a :class:`ModificationRule` instances from a :class:`dict` scraped
        from Unimod.

        Parameters
        ----------
        unimod_entry : dict
            The Unimod entry description for this modificatoin

        Returns
        -------
        :class:`ModificationRule`
        """
        specificity = unimod_entry["specificity"]
        monoisotopic_mass = unimod_entry["mono_mass"]
        full_name = unimod_entry["full_name"]
        record_id = unimod_entry.get("record_id")
        title = unimod_entry["title"]
        alt_names = set(unimod_entry.get('alt_names', ()))
        if record_id is not None:
            alt_names.add("UNIMOD:%d" % (record_id))
        neutral_losses = list()
        for losses in map(NeutralLoss.from_unimod_specificity, specificity):
            neutral_losses.append(losses)
        specificity = list(map(ModificationTarget.from_unimod_specificity, specificity))
        neutral_losses = dict(zip(specificity, neutral_losses))
        return cls(
            specificity, full_name, title, monoisotopic_mass,
            composition=Composition({str(k): int(v) for k, v in unimod_entry['composition'].items()}),
            alt_names=alt_names, neutral_losses=neutral_losses)

    @staticmethod
    def _get_preferred_name(names):
        return min([x for x in names if not x.startswith("UNIMOD:")], key=len)

    def __init__(self, amino_acid_specificity, modification_name,
                 title=None, monoisotopic_mass=None, composition=None,
                 categories=None, alt_names=None, neutral_losses=None,
                 aliases=None, **kwargs):
        if neutral_losses is None:
            neutral_losses = dict()
        if categories is None:
            categories = []
        if aliases is None:
            aliases = set()

        self.mass = float(monoisotopic_mass)
        self.composition = composition
        self.options = kwargs

        if alt_names is None:
            alt_names = set()

        self._configure_names(modification_name, title, alt_names, aliases)
        self._configure_targets(amino_acid_specificity)
        self._configure_categories(categories)

        self.neutral_losses = neutral_losses

    def _configure_names(self, modification_name, title, alt_names, aliases):

        # Attempt to parse the protein prospector name which contains the
        # target in
        try:
            self.common_name = title_cleaner.search(title).groupdict()['name']
        except Exception:
            self.common_name = modification_name

        self.unimod_name = modification_name
        self.title = title if title is not None else modification_name
        self.names = {self.unimod_name, self.common_name, self.title} | alt_names - {''}
        for name in list(self.names):
            self.names.update(name.split(" or "))
        self.aliases = aliases
        preferred_name = self.options.get("preferred_name")
        if preferred_name is not None:
            self.names.add(preferred_name)
            self.preferred_name = self.name = str(preferred_name)
        else:
            self.preferred_name = self.name = str(
                self._get_preferred_name(self.names))
        self._hash = hash(self.name)

    def _configure_categories(self, categories):
        self.categories = list(categories)
        categories = set(self.categories)
        for target in self.targets:
            categories.update(
                target.classification)
        self.categories = list(categories)

    def clone(self, propagated_targets=None):
        if propagated_targets is None:
            propagated_targets = (self.targets)

        dup = self.__class__(
            set(propagated_targets), self.name,
            self.title, self.mass, self.composition.copy(),
            self.categories, **self.options)
        dup.names.update(self.names)
        return dup


    def __repr__(self):
        rep = "{name}:{mass}".format(
            name=self.name, mass=self.mass)
        return rep

    def __add__(self, other):
        if(isinstance(other, ModificationRule)):
            if (other.targets) is (self.targets):
                return self
            dup = self.clone()
            dup.targets = (set(self.targets) | set(other.targets))
            if self.composition is None:
                new_composition = other.composition
            else:
                new_composition = self.composition
            dup.composition = new_composition
            dup.names = self.names | other.names
            dup.aliases = self.aliases | other.aliases
            dup.name = self._get_preferred_name(dup.names)
            dup.neutral_losses = _merge_dictlists(self.neutral_losses, other.neutral_losses)
            dup.options.update(other.options)
            return dup
        else:
            raise TypeError(
                "Unsupported types. Can only add two ModificationRules together.")

    def __sub__(self, other):
        if isinstance(other, ModificationRule):
            dup = self.clone()
            dup.targets = set(self.targets) - set(other.targets)
            return dup
        else:
            raise TypeError(
                "Unsupported types. Can only subtract two ModificationRules together.")

    def __call__(self, **kwargs):
        return self.modification_tp(self, **kwargs)

    def __reduce__(self):
        return _ModificationRule_reconstructor, (self.__class__, ), self.__getstate__()

    def __getstate__(self):
        state = {}
        state['targets'] = self.targets
        state['mass'] = self.mass
        state['composition'] = self.composition
        state['aliases'] = self.aliases
        state['options'] = self.options
        state['categories'] = self.categories
        state['neutral_losses'] = self.neutral_losses
        state['names'] = self.names
        state['unimod_name'] = self.unimod_name
        state['title'] = self.title
        state['name'] = self.name
        state['common_name'] = self.common_name
        state['_n_term_targets'] = self.n_term_targets
        state['_c_term_targets'] = self.c_term_targets
        return state

    def __setstate__(self, state):
        self.targets = state['targets']
        self.mass = state['mass']
        self.composition = state['composition']
        self.aliases = state['aliases']
        self.options = state['options']
        self.categories = state['categories']
        self.names = state['names']
        self.neutral_losses = state['neutral_losses']
        self.title = state['title']
        # defend against outdated pickles here
        self.name = state.get(
            'name', state.get(
                'preferred_name', self._get_preferred_name(
                    state.get('names', []))))
        self.common_name = state['common_name']
        self.unimod_name = state['unimod_name']
        self._n_term_targets = state.get('_n_term_targets', state.get('_n_term_target', None))
        self._c_term_targets = state.get('_c_term_targets', state.get('_c_term_target', None))
        self._hash = hash(self.name)

    def as_spec_strings(self):
        for target in self.targets:
            yield "%s (%s)" % (self.name, target.serialize())

    @property
    def is_core(self):
        return True


@ModificationRule.resolve.register
class AnonymousModificationRule(ModificationRule):
    '''
    Construct a modification rule from a string that can be placed
    in a peptide sequence and be reconstructed without being stored
    in a :class:`ModificationTable` instance and backed by a modification
    database.

    "...(@{name}-{mass})..."

    '''
    parser = re.compile(r"@(?P<name>.+?)-(?P<mass>-?[0-9\.]+)")

    @classmethod
    def try_parse(cls, rule_string):
        anon = cls.parser.search(rule_string)
        if anon is not None:
            anon_match_data = anon.groupdict()
            name = anon_match_data['name']
            mass = float(anon_match_data['mass'])
            mod = AnonymousModificationRule(name, mass)
            return mod
        else:
            return None

    def __init__(self, name, mass):
        try:
            super(AnonymousModificationRule, self).__init__(
                "", name, name, mass)
        except AttributeError:
            pass

    @property
    def is_standard(self):
        return False

    def valid_site(self, *args, **kwargs):  # pylint: disable=arguments-differ
        raise TypeError(
            "AnonymousModificationRule does not support site validation")

    def find_valid_sites(self, *args, **kwargs):  # pylint: disable=arguments-differ
        raise TypeError(
            "AnonymousModificationRule does not support site validation")

    def serialize(self):
        return "@" + self.name + "-" + str(self.mass)

    def clone(self, propagated_targets=None):
        dup = self.__class__(self.name, self.mass, **self.options)
        return dup

    def __repr__(self):
        rep = "{name}:{mass}".format(name=self.name, mass=self.mass)
        return rep


@ModificationRule.resolve.register
class AminoAcidSubstitution(AnonymousModificationRule):
    parser = re.compile(r"@(?P<original_residue>\S+)->(?P<substitution_residue>\S+)")

    @classmethod
    def try_parse(cls, rule_string):
        aa_sub = cls.parser.search(rule_string)
        if aa_sub is not None:
            match_data = aa_sub.groupdict()
            mod = cls(match_data['original_residue'], match_data['substitution_residue'])
            return mod
        else:
            return None

    @classmethod
    def _simple_sequence_mass(cls, seq_list):
        '''Because AminoAcidSubstitution may contain sequence-like entries that
        need massing but we cannot import `sequence` here, this defines a simple
        sequence list to total mass without any terminal groups, but handles residues
        and modifications.

        For a full sequence_to_mass converter, see the sequence module'''
        from .modification import Modification
        mass = 0.0
        for res, mods in seq_list:
            mass += AminoAcidResidue(res).mass
            for mod in mods:
                if mod != "":
                    mass += Modification(mod).mass
        return mass

    def __init__(self, original_residue, substitution_residue, **kwargs): # pylint: disable=super-init-not-called
        if not isinstance(original_residue, AminoAcidResidue):
            original_residue = AminoAcidResidue(original_residue)
        if not isinstance(substitution_residue, AminoAcidResidue):
            substitution_residue = AminoAcidResidue(substitution_residue)
        self.common_name = "{0}->{1}".format(original_residue, substitution_residue)
        self.original = (original_residue)
        self.substitution = (substitution_residue)
        self.mass = (self.substitution).mass - (self.original).mass
        self.name = self.common_name
        self.names = {self.common_name}
        self.options = kwargs
        self.categories = [ModificationCategory['substitution']]
        self.aliases = set()
        self._hash = hash(self.name)

    def serialize(self):
        return "@" + self.name

    @property
    def neutral_losses(self):
        return []

    def __repr__(self):
        return "{name}:{delta}".format(name=self.name, delta=self.mass)

    def __reduce__(self):
        return _ModificationRule_reconstructor, (self.__class__, ), self.__getstate__()

    def __getstate__(self):
        state = {}
        state['mass'] = self.mass
        state['composition'] = self.composition
        state['options'] = self.options
        state['categories'] = self.categories
        state['names'] = self.names
        state['name'] = self.name
        state['common_name'] = self.common_name
        state['original'] = self.original
        state['substitution'] = self.substitution
        return state

    def __setstate__(self, state):
        self.mass = state['mass']
        self.composition = state['composition']
        self.options = state['options']
        self.categories = state['categories']
        self.names = state['names']
        # defend against outdated pickles here
        self.name = state.get(
            'name', state.get(
                'preferred_name', self._get_preferred_name(
                    state.get('names', []))))
        self.common_name = state['common_name']
        self.original = state['original']
        self.substitution = state['substitution']
        self._hash = hash(self.name)



TMT10plex = [
    ('TMT10-126', 125.12044953),
    ('TMT10-127N', 126.1174845),
    ('TMT10-127C', 126.1238045),
    ('TMT10-128N', 127.1208395),
    ('TMT10-128C', 127.1271595),
    ('TMT10-129N', 128.1241945),
    ('TMT10-129C', 128.1305135),
    ('TMT10-130N', 129.1275485),
    ('TMT10-130C', 129.1338685),
    ('TMT10-131', 130.13090353)
]


try:
    from glycopeptidepy._c.structure.modification.rule import NeutralLossBase
except ImportError:
    class NeutralLossBase(object):
        def __eq__(self, other):
            return self.label == other.label and self.mass == other.mass

        def __ne__(self, other):
            return not self == other

        def __hash__(self):
            return hash(self.label)


class NeutralLoss(NeutralLossBase):

    def __init__(self, composition, mass, label=None):
        if label is None:
            label = formula(composition)
        self.composition = composition
        self.mass = mass
        self.label = label

    def __repr__(self):
        template = "{self.__class__.__name__}({self.composition}, {self.mass}, {self.label})"
        return template.format(self=self)

    @classmethod
    def from_unimod_specificity(cls, specificity):
        result = []
        for loss in specificity.get('neutral_losses', []):
            if loss['mono_mass'] == 0:
                continue
            result.append(cls(Composition(loss['composition']), loss['mono_mass']))
        return result



import re

from functools import partial
from typing import List, Dict, Tuple, TYPE_CHECKING, Optional, Set

from six import string_types as basestring

from glypy import Substituent, Glycan
from glypy.io import glycoct, iupac, linear_code, wurcs
from glypy.structure.glycan_composition import (
    FrozenMonosaccharideResidue, FrozenGlycanComposition)


from ..composition import Composition
from ..glycan import TypedGlycan, TypedGlycanComposition, GlycosylationType

from .target import ModificationTarget
from .rule import ModificationRule, ModificationLocatingMixin
from .descriptors import ModificationCategory


_hexnac = FrozenMonosaccharideResidue.from_iupac_lite("HexNAc")
_hexose = FrozenMonosaccharideResidue.from_iupac_lite("Hex")
_xylose = FrozenMonosaccharideResidue.from_iupac_lite("Xyl")

if TYPE_CHECKING:
    from glycopeptidepy.structure import PeptideSequence


hexnac_modification = ModificationRule.from_unimod({
    "title": "HexNAc",
    "composition": {
        "H": 13,
        "C": 8,
        "O": 5,
        "N": 1
    },
    "record_id": 43,
    "mono_mass": 203.079373,
    "full_name": "N-Acetylhexosamine",
    "specificity": [{
        "position": "Anywhere",
        "hidden": True,
        "site": "T",
        "classification": "Other glycosylation",
        "spec_group": 3
    }, {
        "position": "Anywhere",
        "hidden": True,
        "site": "S",
        "classification": "Other glycosylation",
        "spec_group": 2
    }, {
        "position": "Anywhere",
        "hidden": True,
        "site": "N",
        "classification": "N-linked glycosylation",
        "spec_group": 1
    }],
})


xylose_modification = ModificationRule.from_unimod({
    "title": "Xyl",
    "composition": dict(_xylose.total_composition()),
    "mono_mass": _xylose.mass(),
    "full_name": "Xylose",
    "specificity": [{
        "position": "Anywhere",
        "hidden": True,
        "site": "S",
        "classification": "O-linked glycosylation",
        "spec_group": 2
    }],
})


glycan_resolvers = dict()
glycan_resolvers['glycoct'] = partial(glycoct.loads, structure_class=TypedGlycan)
glycan_resolvers['iupac'] = partial(iupac.loads, structure_class=TypedGlycan)
glycan_resolvers['iupac_simple'] = partial(iupac.loads, structure_class=TypedGlycan, dialect='simple')
glycan_resolvers['linear_code'] = partial(linear_code.loads, structure_class=TypedGlycan)
glycan_resolvers['wurcs2'] = partial(wurcs.loads, structure_class=TypedGlycan)
glycan_resolvers['iupaclite'] = TypedGlycanComposition.parse


def parse_glycan(glycan_format, encoded_string):
    try:
        parser = glycan_resolvers[glycan_format]
        result = parser(encoded_string)
        if not result:
            raise KeyError(encoded_string)
        return result
    except KeyError:
        raise KeyError("Could not resolve glycan parser for %r (%r)" % (glycan_format, encoded_string))


def _Glycosylation_reconstructor(string):
    return Glycosylation.try_parse(string)


class GlycosylationBase(ModificationRule):

    @property
    def glycosylation_type(self):
        return self._glycosylation_type

    @property
    def is_composition(self):
        return self._is_composition

    @property
    def is_core(self):
        return self._is_core

    def is_tracked_for(self, category):
        return category == ModificationCategory.glycosylation


try:
    from glycopeptidepy._c.structure.modification.rule import GlycosylationBase
except ImportError:
    pass


@ModificationRule.resolve.register
class Glycosylation(GlycosylationBase, ModificationLocatingMixin):
    """
    Incubator Idea - Represent occupied glycosylation sites
    using the Modification interface.

    Attributes
    ----------
    categories : list
        Description
    mass : float
        The mass of the glycosylation modification
    name : str
        Description
    names : set[str]
        Description
    options : dict
        Description
    """


    @classmethod
    def _parse(cls, rule_string):
        if rule_string.startswith("#"):
            rule_string = rule_string[1:]
        format_type = "iupaclite"
        glycan_definition = rule_string
        metadata = {}
        if rule_string.startswith(":"):
            match = re.search(r":([^:]*?):(.+)", rule_string, re.DOTALL)
            if match:
                metadata = match.group(1).split(",")
                if '=' not in metadata[0]:
                    format_type = metadata[0]
                    if not format_type:
                        format_type = 'iupaclite'
                    metadata = metadata[1:]
                metadata = dict([token.split("=") for token in metadata])
                glycan_definition = match.group(2)
            else:
                raise ValueError("Cannot recognize glycan format %r" % (rule_string,))

        return parse_glycan(format_type, glycan_definition), format_type, metadata

    @classmethod
    def try_parse(cls, rule_string):
        try:
            glycan, encoding_format, metadata = cls._parse(rule_string)
            return cls(glycan, encoding_format, metadata)
        except Exception:
            return None

    def __init__(self, glycan, encoding_format=None, metadata=None):
        if metadata is None:
            metadata = {}
        if isinstance(glycan, basestring):
            if encoding_format is None:
                glycan, encoding_format, _metadata = self._parse(glycan)
                metadata.update(_metadata)
            else:
                glycan = parse_glycan(encoding_format, glycan)
        else:
            glycan = glycan.clone()

        self._is_core = False

        if isinstance(glycan, Glycan):
            self._is_composition = False
            if encoding_format is None:
                encoding_format = "glycoct"
        else:
            self._is_composition = True
            encoding_format = "iupaclite"

        # Prepare information to encode the rule
        self.encoding_format = encoding_format
        self.metadata = metadata
        self.glycan = glycan
        try:
            self._glycosylation_type = self.glycan.glycosylation_type = GlycosylationType[metadata['glycosylation_type']]
        except KeyError:
            pass
        self._original = glycan.clone()
        self.common_name = self._make_string()

        # Only after the common name is set, prepare the patch
        self._patch_dehydration()
        self.mass = self.glycan.mass()
        self.composition = self.glycan.total_composition()

        self._simple_configuration()

    def __reduce__(self):
        return _Glycosylation_reconstructor, (self.name, )

    def _simple_configuration(self):
        self.title = self.common_name
        self.name = self.common_name
        self.unimod_name = self.common_name
        self.categories = [ModificationCategory.glycosylation]
        self.names = {self.common_name, self.title, self.name}
        self.aliases = set()
        self.options = {}
        self._hash = hash(self.name)
        self.neutral_losses = []
        self._n_term_targets = None
        self._c_term_targets = None
        self.targets = set()

    def _make_string(self):
        template = "#:{}:{}"
        metadata = ",".join("%s=%s" % (k, v) for k, v in self.metadata.items())
        if metadata:
            if self.encoding_format != 'iupaclite':
                metadata = "{},{}".format(self.encoding_format, metadata)
        else:
            if self.encoding_format != 'iupaclite':
                metadata = self.encoding_format
        if self.is_composition:
            return template.format(metadata, self.glycan.serialize())
        else:
            return template.format(
                metadata, self.glycan.serialize(self.encoding_format))

    def _patch_dehydration(self):
        if self.is_composition:
            self.glycan.composition_offset = Composition()
        else:
            self.glycan.root.add_substituent(
                Substituent("aglycone", composition=Composition()),
                position=1, parent_loss=Composition("HO"))

    def clone(self):
        return self.__class__(
            self._original, self.encoding_format, self.metadata)

    def get_fragments(self, *args, **kwargs):
        if self.is_composition:
            raise TypeError("Cannot generate fragments from composition")
        for frag in self.glycan.fragments(*args, **kwargs):
            yield frag

    def total_composition(self):
        return self.glycan.total_composition()

    @staticmethod
    def modification_tp(*args, **kwargs):
        from .modification import Modification
        return Modification(*args, **kwargs)

    def __call__(self, **kwargs):
        return self.modification_tp(self, **kwargs)


class CoreGlycosylation(Glycosylation):
    def get_fragments(self, *args, **kwargs):
        return []

    def _common_init(self):
        self.names = {self.unimod_name, self.title, self.name, self.common_name}
        self.options = {}
        self.aliases = set()
        self.categories = [ModificationCategory.glycosylation]
        self._hash = hash(self.name)
        self.neutral_losses = []
        self._n_term_targets = None
        self._c_term_targets = None
        self._is_composition = True
        self._is_core = True
        self.targets = set()

    def clone(self):
        return self.__class__(self.mass, )

    def __reduce__(self):
        return self.__class__, (self.mass, )

    def __repr__(self):
        return self.name


class NGlycanCoreGlycosylation(CoreGlycosylation):
    mass_ladder = {k: FrozenGlycanComposition.parse(k).total_composition() - Composition("H2O") for k in {
        "{HexNAc:1}",
        "{HexNAc:2}",
        "{HexNAc:2; Hex:1}",
        "{HexNAc:2; Hex:2}",
        "{HexNAc:2; Hex:3}",
    }}

    def __init__(self, base_mass=_hexnac.mass()):
        self.common_name = "NGlycanCoreGlycosylation"
        self.mass = base_mass
        self.title = "N-Glycan Core Glycosylation"
        self.unimod_name = "N-Glycosylation"
        self.name = self.unimod_name
        self.composition = _hexnac.total_composition().clone()
        self._glycosylation_type = GlycosylationType.n_linked
        self._common_init()
        self.targets = {ModificationTarget("N")}

    def find_valid_sites(self, sequence: 'PeptideSequence', protein_n_term=False, protein_c_term=False):
        from glycopeptidepy.structure.sequence import find_n_glycosylation_sequons
        valid = set(find_n_glycosylation_sequons(sequence, include_cysteine=True))
        n = len(sequence)
        any_valid = set(super().find_valid_sites(sequence, protein_n_term, protein_c_term))
        for i in any_valid:
            if (n - i) < 3:
                valid.add(i)
        return sorted(valid)


class MucinOGlycanCoreGlycosylation(CoreGlycosylation):
    mass_ladder = {k: FrozenGlycanComposition.parse(k).total_composition() - Composition("H2O") for k in {
        "{HexNAc:1}",
        "{HexNAc:1; Hex:1}",
    }}

    def __init__(self, base_mass=_hexnac.mass()):
        self.common_name = "MucinOGlycanCoreGlycosylation"
        self.mass = base_mass
        self.title = "Mucin O-Glycan Core Glycosylation"
        self.unimod_name = "O-Glycosylation"
        self.name = self.unimod_name
        self.composition = _hexnac.total_composition().clone()
        self._glycosylation_type = GlycosylationType.o_linked
        self._common_init()
        self.targets = {ModificationTarget("S"), ModificationTarget("T")}


OGlycanCoreGlycosylation = MucinOGlycanCoreGlycosylation


class GlycosaminoglycanLinkerGlycosylation(CoreGlycosylation):
    mass_ladder = {k: FrozenGlycanComposition.parse(k).total_composition() - Composition("H2O") for k in {
        "{Xyl:1}",
        "{Xyl:1; Hex:1}",
        "{Xyl:1; Hex:2}",
        "{Xyl:1; Hex:2; HexA:1}",
    }}

    def __init__(self, base_mass=_xylose.mass()):
        self.common_name = "GlycosaminoglycanLinkerGlycosylation"
        self.mass = base_mass
        self.title = "Glycosaminoglycan Linker Glycosylation"
        self.unimod_name = "GAG-Linker"
        self.name = self.unimod_name
        self.composition = _hexnac.total_composition().clone()
        self._glycosylation_type = GlycosylationType.glycosaminoglycan
        self._common_init()
        self.targets = {ModificationTarget("S")}

class OGlcNAcylation(CoreGlycosylation):
    mass_ladder = {
        "{GlcNAc:1}": _hexnac.total_composition()
    }

    def __init__(self, base_mass=_hexnac.mass()):
        self.common_name = "O-GlcNAc"
        self.name = self.common_name
        self.mass = base_mass
        self.title = self.common_name
        self.unimod_name = self.common_name
        self.targets = {ModificationTarget("S"), ModificationTarget("T")}
        self.composition = _hexnac.total_composition().clone()
        self._common_init()

    def get_fragments(self, *args, **kwargs):
        for label_loss in self.mass_ladder.items():
            yield label_loss


def _GlycanFragment_reconstructor():
    return GlycanFragment.__new__(GlycanFragment)


class GlycanFragment(Glycosylation):
    def get_fragments(self, *args, **kwargs):
        return []

    def __init__(self, fragment):
        self.mass = fragment.mass
        self.name = self.unimod_name = self.title = self.common_name = fragment.name
        self.names = {self.name, }
        self.composition = fragment.composition
        self.fragment = fragment
        self._set_defaults()

    def _set_defaults(self):
        self._is_core = False
        self._is_composition = True
        self.categories = [ModificationCategory.glycosylation]
        self.aliases = set()
        self.targets = set()
        self.options = {}
        self._hash = hash(self.name)


    def __reduce__(self):
        return _GlycanFragment_reconstructor, (), self.__getstate__()

    def __getstate__(self):
        state = {
            'name': self.name,
            'composition': self.composition,
            'mass': self.mass,
            "fragment": self.fragment
        }
        return state

    def __setstate__(self, state):
        self.mass = state['mass']
        self.name = self.unimod_name = self.title = self.common_name = state['name']
        self.names = {self.name, }
        self.composition = state['composition']
        self.fragment = state.get("fragment")
        self._set_defaults()

    def __repr__(self):
        return self.name



cimport cython
from cpython.object cimport PyObject
from cpython.list cimport PyList_Append, PyList_Size
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem, PyDict_Next, PyDict_Size

from glypy.composition.ccomposition cimport CComposition
from glypy._c.structure.glycan_composition cimport _CompositionBase
from glypy.structure.glycan_composition import HashableGlycanComposition

from glycopeptidepy._c.structure.modification.modification cimport ModificationInstanceBase
from glycopeptidepy._c.structure.modification.rule cimport ModificationRuleBase, GlycosylationBase
from glycopeptidepy._c.structure.base cimport PeptideSequenceBase


@cython.freelist(10000)
cdef class GlycanCompositionProxy(object):
    '''A mapping-like object that imitates the GlycanComposition interface in
    a read-only fashion.
    '''

    @staticmethod
    cdef GlycanCompositionProxy _create(glycan_composition_type obj):
        cdef:
            GlycanCompositionProxy self
        self = <GlycanCompositionProxy>GlycanCompositionProxy.__new__(GlycanCompositionProxy)
        if glycan_composition_type is _CompositionBase:
            self.obj = obj
            self._serialized = None
        elif glycan_composition_type is GlycanCompositionProxy:
            self.obj = obj.obj
            self._serialized = obj._serialized
        return self

    def __init__(self, glycan_composition):
        if isinstance(glycan_composition, GlycanCompositionProxy):
            self._serialized = glycan_composition._serialized
            self.obj = glycan_composition.obj
        else:
            self.obj = glycan_composition
            self._serialized = None

    def mass(self, *args, **kwargs):
        return self.obj.mass(*args, **kwargs)

    def total_composition(self, *args, **kwargs):
        return self.obj.total_composition(*args, **kwargs)

    def __iter__(self):
        return iter(self.obj)

    def __getitem__(self, key):
        return self.obj[key]

    cpdef object _getitem_fast(self, key):
        return self.obj._getitem_fast(key)

    def __contains__(self, key):
        return key in self.obj

    def keys(self):
        return self.obj.keys()

    def values(self):
        return self.obj.values()

    def items(self):
        return self.obj.items()

    def clone(self, *args, **kwargs):
        return self.obj.clone(*args, **kwargs)

    def query(self, query, exact=True, **kwargs):
        return self.obj.query(query, exact=exact, **kwargs)

    cpdef str serialize(self):
        if self._serialized is not None:
            return self._serialized
        self._serialized = self.obj.serialize()
        return self._serialized

    def __repr__(self):
        return repr(self.obj)

    def __str__(self):
        if self._serialized is not None:
            return self._serialized
        return self.serialize()

    def __hash__(self):
        return hash(self.obj)

    def __eq__(self, other):
        try:
            other = other.obj
        except AttributeError:
            pass
        return self.obj == other

    def __ne__(self, other):
        try:
            other = other.obj
        except AttributeError:
            pass
        return (self.obj != other)

    def __add__(self, other):
        return self.obj + other

    def __sub__(self, other):
        return self.obj - other

    def __mul__(self, other):
        return self.obj * other

    def __len__(self):
        return len(self.obj)


cdef class GlycanCompositionWithOffsetProxy(GlycanCompositionProxy):
    def __init__(self, obj, offset=None):
        if offset is None:
            offset = CComposition._create(None)
        GlycanCompositionProxy.__init__(self, obj)
        self.composition_offset = offset


    def clone(self, *args, **kwargs):
        return self.__class__(self.obj, self.composition_offset.copy())

    def mass(self, *args, **kwargs):
        mass = self.obj.mass(*args, **kwargs)
        return mass + self.composition_offset.mass

    def total_composition(self):
        comp = self.obj.total_composition()
        return comp + self.composition_offset


cdef CComposition WATER_OFFSET = CComposition({"H": 2, "O": 1})


cdef object GlycanCompositionProxyType = None


@cython.freelist(10000)
cdef class GlycosylationManager(object):
    def __init__(self, parent, aggregate=None):
        self._init()
        self.parent = parent
        if aggregate is not None:
            self.set_aggregate(aggregate)

    @staticmethod
    cdef GlycosylationManager _create(PeptideSequenceBase parent, object aggregate):
        cdef:
            GlycosylationManager self
        self = GlycosylationManager.__new__(GlycosylationManager)
        self._init()
        self.parent = parent
        if aggregate is not None:
            self.set_aggregate(aggregate)
        return self

    cdef void _init(self):
        self.mapping = {}
        self._aggregate = None
        self._proxy = None
        self._type_track = None
        self._total_glycosylation_size = -1

    def __getitem__(self, key):
        return self.mapping[key]

    def __setitem__(self, key, value):
        self.invalidate()
        self.mapping[key] = value

    def __delitem__(self, key):
        self.invalidate()
        del self.mapping[key]

    def __len__(self):
        return self.get_size()

    cdef size_t get_size(self):
        return PyDict_Size(self.mapping)

    cpdef keys(self):
        return self.mapping.keys()

    cpdef values(self):
        return self.mapping.values()

    cpdef items(self):
        return self.mapping.items()

    cpdef pop(self, key, default=None):
        self.invalidate()
        self.mapping.pop(key, default)

    cpdef clear(self):
        self.invalidate()
        self.set_aggregate(None)
        self.mapping.clear()

    cpdef update(self, other):
        self.invalidate()
        if isinstance(other, GlycosylationManager):
            self.mapping.update(other.mapping)
        else:
            self.mapping.update(other)

    cpdef GlycosylationManager copy(self):
        cdef:
            GlycosylationManager inst
            object aggregate
        aggregate = self.get_aggregate()

        inst = GlycosylationManager._create(
            self.parent,
            aggregate.clone() if aggregate is not None else None)
        inst.update(self)
        return inst

    cpdef GlycosylationManager clone(self):
        return self.copy()

    cpdef object _patch_aggregate(self):
        cdef:
            object aggregate
            CComposition offset
        aggregate = self.get_aggregate()
        if isinstance(aggregate, GlycanCompositionWithOffsetProxy):
            offset = (<GlycanCompositionWithOffsetProxy>aggregate).composition_offset
        elif isinstance(aggregate, _CompositionBase):
            offset = (<_CompositionBase>aggregate)._composition_offset
        else:
            offset = <CComposition>(aggregate).composition_offset
        offset.subtract_from(WATER_OFFSET)

    cpdef object invalidate(self):
        self._proxy = None
        self._type_track = None
        self._total_glycosylation_size = -1

    cdef object get_aggregate(self):
        return self._aggregate

    cdef int set_aggregate(self, object value):
        self._aggregate = value
        if self._aggregate is not None:
            self._patch_aggregate()
        self.invalidate()
        return 0

    @property
    def aggregate(self):
        return self.get_aggregate()

    @aggregate.setter
    def aggregate(self, value):
        self.set_aggregate(value)

    cdef dict get_glycosylation_types(self):
        if self._type_track is None:
            self._type_track = self._update_track()
        return self._type_track

    cdef dict _update_track(self):
        cdef:
            dict track
            Py_ssize_t pos
            PyObject* pkey
            PyObject* pval
            PyObject* tmp
            list bucket
            GlycosylationBase glycosylation

        track = dict()
        pos = 0
        while PyDict_Next(self.mapping, &pos, &pkey, &pval):
            glycosylation = <GlycosylationBase>(<ModificationInstanceBase>pval).rule
            glycosylation_type = glycosylation.get_glycosylation_type()
            tmp = PyDict_GetItem(track, glycosylation_type)
            if tmp == NULL:
                bucket = []
                PyDict_SetItem(track, glycosylation_type, bucket)
            else:
                bucket = <list>tmp
            PyList_Append(bucket, (<object>pkey, <object>pval))
        return  track

    cpdef CComposition total_composition(self):
        cdef:
            CComposition total
            bint has_aggregate
            PyObject* pkey
            PyObject* pval
            Py_ssize_t pos
            ModificationInstanceBase value
            object aggregate
        aggregate = self.get_aggregate()
        total = CComposition._create(None)
        has_aggregate = aggregate is not None
        pos = 0
        while PyDict_Next(self.mapping, &pos, &pkey, &pval):
            value = <ModificationInstanceBase>pval
            if has_aggregate and (<GlycosylationBase>(value.rule)).is_core:
                continue
            total.add_from(value.composition)
        if has_aggregate:
            total.add_from(aggregate.total_composition())
        return total

    @property
    def glycan_composition(self):
        return self.get_glycan_composition()

    cdef GlycanCompositionProxy get_glycan_composition(self):
        if self._proxy is None:
            self._proxy = self._make_glycan_composition_proxy()
        return self._proxy

    cpdef double mass(self):
        cdef:
            double total
            bint has_aggregate
            PyObject* pkey
            PyObject* pval
            Py_ssize_t pos
            ModificationInstanceBase value
            object aggregate
        total = 0.
        aggregate = self.get_aggregate()
        has_aggregate = aggregate is not None
        pos = 0
        while PyDict_Next(self.mapping, &pos, &pkey, &pval):
            value = <ModificationInstanceBase>pval
            if has_aggregate and (<GlycosylationBase>(value.rule)).is_core:
                continue
            total += value.mass
        if has_aggregate:
            total += aggregate.mass()
        return total

    cdef int get_total_glycosylation_size(self):
        if self._total_glycosylation_size == -1:
            aggregate = self.get_aggregate()
            if aggregate is None:
                self._total_glycosylation_size = 0
            else:
                self._total_glycosylation_size = sum(self.get_aggregate().values())
        return self._total_glycosylation_size

    cpdef int total_glycosylation_size(self):
        return self.get_total_glycosylation_size()

    cpdef bint is_fully_specified_topologies(self):
        cdef:
            bint is_fully_specified
            PyObject* pkey
            PyObject* pval
            Py_ssize_t pos
            ModificationInstanceBase value
            GlycosylationBase rule

        is_fully_specified = self.get_size() > 0
        pos = 0
        while PyDict_Next(self.mapping, &pos, &pkey, &pval):
            value = <ModificationInstanceBase>pval
            rule = <GlycosylationBase>value.rule
            if rule.is_core:
                is_fully_specified = False
                break
            elif rule.is_composition:
                is_fully_specified = False
                break
        return is_fully_specified

    cdef GlycanCompositionProxy _make_glycan_composition_proxy(self):
        cdef:
            object aggregate
            object base
            ModificationInstanceBase value
            GlycosylationBase rule
            Py_ssize_t pos
            PyObject* pkey
            PyObject* pval

        aggregate = self.get_aggregate()
        if aggregate is not None:
            base = aggregate.clone()
        else:
            base = HashableGlycanComposition()
            # Represent the initial amide bond between the peptide
            # and the first glycan. Subsequent glycans do not need
            # further chemical losses because of the dehyration built
            # directly into the Residue abstraction.
            base.composition_offset -= WATER_OFFSET
        pos = 0
        while PyDict_Next(self.mapping, &pos, &pkey, &pval):
            value = <ModificationInstanceBase>pval
            rule = <GlycosylationBase>(value.rule)
            if rule.is_core:
                continue
            elif rule.is_composition:
                base += rule.glycan
            else:
                # Convert Glycan object into a composition, using the original
                # detatched topology to omit the "aglycone" group which represents
                # the connection between the glycan and the peptide, which penalizes
                # the composition by H2O. This H2O is lost when that bond is formed,
                # but doesn't need to be explicitly included as the loss is tracked
                # when initializing the base above.
                gc = HashableGlycanComposition.from_glycan(rule._original)
                base += gc
        if not isinstance(base, GlycanCompositionProxy):
            proxy = GlycanCompositionProxy(base)
        else:
            proxy = base
        return proxy

    cpdef int count_glycosylation_type(self, glycosylation_type):
        cdef:
            dict tracks
            PyObject* tmp
        tracks = self.get_glycosylation_types()
        tmp = PyDict_GetItem(tracks, glycosylation_type)
        if tmp == NULL:
            return 0
        else:
            return PyList_Size(<list>tmp)

    def __repr__(self):
        return "{self.__class__.__name__}({self.mapping})".format(self=self)


from cpython.object cimport PyObject
from cpython.ref cimport Py_INCREF, Py_DECREF
from cpython.iterator cimport PyIter_Next
from cpython.int cimport PyInt_AsLong
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem, PyDict_Next, PyDict_Keys, PyDict_Values
from cpython.list cimport PyList_Append, PyList_GetItem, PyList_Size, PyList_SetItem
from cpython.tuple cimport PyTuple_GetItem, PyTuple_Size
from cpython.set cimport PySet_Size

import itertools

cdef object product = itertools.product


cdef class LimitedCrossproduct(object):
    cdef:
        public list collections
        public size_t max_depth
        public size_t size
        public list precomputed
        public object iterator

    def __init__(self, collections, max_depth=4):
        self.collections = [list(v) for v in collections]
        self.max_depth = PyInt_AsLong(max_depth)
        self.size = PyList_Size(self.collections)

        self.precomputed = self.compose_iterative()
        self.iterator = iter(self.precomputed)

    def compose(self):
        i = 0
        current = []
        depth = 0
        for res in self.compose_inner(i, current, depth):
            yield res

    def compose_inner(self, layer_index, current, depth):
        if layer_index == self.size:
            yield current, depth
            return
        layer = self.collections[layer_index]
        for val in layer:
            new_depth = depth + (val is not None)
            if new_depth <= self.max_depth:
                for res in self.compose_inner(layer_index + 1, current + [val], new_depth):
                    yield res
            else:
                continue

    cdef list compose_iterative(self):
        cdef:
            list previous, current, layer, rec
            tuple rec_depth
            size_t depth, new_depth
            size_t layers_i, layers_n
            size_t previous_i, previous_n
            size_t current_layer_i, current_layer_n

        previous = [([], 0)]
        layers_i = 0
        layers_n = self.size
        for layers_i in range(layers_n):
            current = []
            layer = <list>PyList_GetItem(self.collections, layers_i)
            current_layer_n = PyList_Size(layer)
            previous_n = PyList_Size(previous)
            for previous_i in range(previous_n):
                rec_depth = <tuple>PyList_GetItem(previous, previous_i)
                rec = <list>PyTuple_GetItem(rec_depth, 0)
                depth = PyInt_AsLong(<object>PyTuple_GetItem(rec_depth, 1))
                for current_layer_i in range(current_layer_n):
                    val = <object>PyList_GetItem(layer, current_layer_i)
                    new_depth = depth + (val is not None)
                    if new_depth <= self.max_depth:
                        current.append((rec + [val], new_depth))
            previous = current
        return previous

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iterator)


cdef class ModificationSiteAssignmentCombinator(object):
    cdef:
        public dict modification_to_site
        public dict site_to_modification
        public size_t max_modifications

    def __init__(self, variable_site_map, max_modifications=4):
        self.modification_to_site = variable_site_map
        self.site_to_modification = self.transpose_sites()
        self.max_modifications = PyInt_AsLong(max_modifications)

    cdef dict transpose_sites(self):
        """Given a dictionary mapping between modification names and
        an iterable of valid sites, create a dictionary mapping between
        modification names and a list of valid sites plus the constant `None`

        Returns
        -------
        dict
        """
        cdef:
            dict sites
            PyObject* pkey
            PyObject* pval
            PyObject* tmp
            Py_ssize_t pos
            object mod, site
            set varsites
            list bucket
            size_t i, n

        sites = dict()
        pos = 0
        while PyDict_Next(self.modification_to_site, &pos, &pkey, &pval):
            mod = <object>pkey
            varsites = <set>pval
            for site in varsites:
                tmp = PyDict_GetItem(sites, site)
                if tmp == NULL:
                    bucket = []
                    PyDict_SetItem(sites, site, bucket)
                else:
                    bucket = <list>tmp
                PyList_Append(bucket, mod)

        pos = 0
        while PyDict_Next(sites, &pos, &pkey, &pval):
            bucket = <list>pval
            PyList_Append(bucket, None)

        return sites


    cdef list _remove_empty_sites(self, list sites, list selected):
        cdef:
            size_t i, n
            list result
            tuple pair
            object p1, p2
        n = PyList_Size(selected)
        result = []
        for i in range(n):
            p2 = <object>PyList_GetItem(selected, i)
            if p2 is None:
                continue
            p1 = <object>PyList_GetItem(sites, i)
            pair = (p1, p2)
            PyList_Append(result, pair)
        return result

    def assign(self):
        cdef:
            list sites
            list choices
            list assigned
            object iterator
            tuple selected
            object tmp
        sites = PyDict_Keys(self.site_to_modification)
        choices = PyDict_Values(self.site_to_modification)

        combinator = LimitedCrossproduct(choices, self.max_modifications)
        n = PyList_Size(combinator.precomputed)
        for i in range(n):
            selected = <tuple>PyList_GetItem(combinator.precomputed, i)
            assigned = <list>PyTuple_GetItem(selected, 0)
            assigned = self._remove_empty_sites(sites, assigned)
            yield assigned

        # iterator = product(*choices)

        # while True:
        #     tmp = PyIter_Next(iterator)
        #     if (<PyObject*>tmp) == NULL:
        #         break
        #     else:
        #         selected = <tuple>tmp
        #         assigned = self._remove_empty_sites(sites, selected)
        #         if PyList_Size(assigned) > self.max_modifications:
        #             continue
        #         yield assigned

        # for selected in product(*choices):
        #     yield self._remove_empty_sites(sites, selected)

    def __iter__(self):
        return self.assign()


import unittest
import pickle

from glycopeptidepy.structure import sequence, modification, residue, composition, parser
from glypy import GlycanComposition, MonosaccharideResidue


HYDROGEN = composition.Composition("H").mass
R = residue.Residue
Modification = modification.Modification

p1 = "PEPTIDE"
p2 = "YPVLN(N-Glycosylation)VTMPN(Deamidation)NGKFDK{Hex:9; HexNAc:2}"
p3 = "NEEYN(N-Glycosylation)K{Hex:5; HexNAc:4; NeuAc:2}"
p4 = "TVDGT(O-Glycosylation)AR{Fuc:1; Hex:1; HexNAc:1; Neu5Ac:1}"
p5 = "(Carbamidomethyl)-FFYFTPNK"
p6 = "NEEYN(N-Glycosylation)K{Fuc:1; Hex:5; HexNAc:4; NeuAc:2}"
p7 = 'ISASGVEDIS(GAG-Linker)R{Xyl:1; a,enHex:1; aHex:1; Hex:1; HexS:1; HexNAc(S):1}'
p8 = ('QQQHLFGSN(#:iupac,glycosylation_type=n_linked:b-D-Galp-(1-4)-b-D-Glcp2NAc-(1-2)'
      '-a-D-Manp-(1-6)-[b-D-Galp-(1-4)-b-D-Glcp2NAc-(1-4)-a-D-Manp-(1-3)]b-D-Manp-(1-4)'
      '-b-D-Glcp2NAc-(1-4)-?-D-Glcp2NAc)VTDC(Carbamidomethyl)SGNFCLFR')
p9 = "N(N-Glycosylation)ITEIVYLTN(N-Glycosylation)TTIEK{Hex:10; HexNAc:8}"
p10 = ('N(#:iupac,glycosylation_type=n_linked:b-D-Galp-(1-4)-b-D-Glcp2NAc-(1-2)-a-D-Manp'
       '-(1-6)-[b-D-Galp-(1-4)-b-D-Glcp2NAc-(1-4)-a-D-Manp-(1-3)]b-D-Manp-(1-4)-b-D-Glcp2NAc'
       '-(1-4)-?-D-Glcp2NAc)ITEIVYLTN(#:iupac,glycosylation_type=n_linked:b-D-Galp-(1-4)'
       '-b-D-Glcp2NAc-(1-2)-a-D-Manp-(1-6)-[b-D-Galp-(1-4)-b-D-Glcp2NAc-(1-4)-a-D-Manp-(1-3)]'
       'b-D-Manp-(1-4)-b-D-Glcp2NAc-(1-4)-?-D-Glcp2NAc)TTIEK')
p11 = ("N(#:glycosylation_type=n_linked:{Hex:5; HexNAc:4})ITEI"
       "VYLTN(#:glycosylation_type=n_linked:{Hex:5; HexNAc:4})TTIEK")
hexnac_mass = MonosaccharideResidue.from_iupac_lite("HexNAc").mass()
hexose_mass = MonosaccharideResidue.from_iupac_lite("Hex").mass()


def make_fragment(self, key):
    for group in self.get_fragments(key[0], compute_compositions=True):
        for frag in group:
            if frag.name == key:
                return frag


class TestSequenceParser(unittest.TestCase):
    def test_parser(self):
        chunks, mods, glycan, n_term, c_term = parser.sequence_tokenizer(p1)
        # self.assertEqual(len(mods), 0)
        self.assertEqual(len(chunks), len(p1))
        self.assertEqual(glycan, None)

        chunks, mods, glycan, n_term, c_term = parser.sequence_tokenizer(p2)
        self.assertEqual(GlycanComposition.parse("{Hex:9; HexNAc:2}"), glycan)
        # self.assertEqual(len(mods), 2)
        self.assertEqual(len(chunks), 16)


class PeptideSequenceSuiteBase(object):
    def parse_sequence(self, seqstr):
        raise NotImplementedError()

    def test_fragmentation(self):
        seq = self.parse_sequence(p1)
        mapping = {
            "b3": 324.1554 - HYDROGEN,
            "b6": 653.3141 - HYDROGEN,
            "y2": 263.0874 - HYDROGEN,
            "y5": 574.2719 - HYDROGEN,
            "a4": 397.2082 - HYDROGEN,
            "a5": 510.2922 - HYDROGEN,
            "z3": 360.1527 - HYDROGEN,
        }

        for fragment, mass in mapping.items():
            fmass = make_fragment(seq, fragment).mass
            self.assertAlmostEqual(mass, fmass, 2, (mass, fmass, fragment))
            self.assertAlmostEqual(
                mass, make_fragment(seq, fragment).total_composition().mass, 2, (mass, fmass, fragment))

    def test_stub_ions(self):
        peptide = self.parse_sequence(p3)
        composition = peptide.total_composition()
        stubs = sorted({f.mass for f in peptide.stub_fragments()})
        self.assertAlmostEqual(stubs[0], 795.3399, 3)
        self.assertAlmostEqual(stubs[1], 998.4193, 3)
        self.assertAlmostEqual(stubs[2], 1201.4986, 3)
        self.assertAlmostEqual(stubs[3], 1363.5515, 3)
        self.assertAlmostEqual(stubs[4], 1525.6043, 3)
        self.assertAlmostEqual(stubs[5], 1687.6571, 3)
        self.assertEqual(composition, peptide.total_composition())

    def test_stub_ions_extended(self):
        peptide = self.parse_sequence(p3)
        stubs = sorted({f for f in peptide.stub_fragments(extended=True)
                        if f.glycosylation}, key=lambda x: x.glycosylation.mass())
        for stub in stubs:
            if sum(stub.glycosylation.values()) > 5:
                assert stub.is_extended
            else:
                assert not stub.is_extended

    def test_glycan_fragments_stubs(self):
        peptide = self.parse_sequence(p3)
        stubs = {f.name: f.mass for f in peptide.glycan_fragments(
            all_series=True, allow_ambiguous=True)}
        self.assertAlmostEqual(
            stubs["peptide+{Hex:3; HexNAc:2}"], 1687.6571, 3)
        self.assertAlmostEqual(stubs["peptide+{Hex:3; HexNAc:3}"],
                               1687.6571 + hexnac_mass, 3)
        self.assertAlmostEqual(stubs["peptide+{Hex:4; HexNAc:3}"],
                               1687.6571 + hexnac_mass + hexose_mass, 3)

    def test_clone(self):
        peptide = self.parse_sequence(p3)
        self.assertEqual(peptide, peptide.clone())
        self.assertEqual(peptide.total_mass, peptide.clone().total_mass)
        self.assertEqual(peptide.total_composition(),
                         peptide.clone().total_composition())
        self.assertTrue(peptide.full_structure_equality(peptide.clone()))
        peptide = self.parse_sequence(p8)
        self.assertEqual(peptide, peptide.clone())
        self.assertEqual(peptide.total_mass, peptide.clone().total_mass)
        self.assertEqual(peptide.total_composition(),
                         peptide.clone().total_composition())
        self.assertTrue(peptide.full_structure_equality(peptide.clone()))

    def test_add_remove_modification(self):
        p = self.parse_sequence("ENGTISR")
        ref_mass = p.mass
        assert not p.has_modification(1, "Deamidated")
        p.add_modification(1, "Deamidated")
        assert p.has_modification(1, "Deamidated")
        self.assertAlmostEqual(self.parse_sequence(str(p)).mass, p.mass, 5)

        p.drop_modification(1, "Deamidated")
        assert not p.has_modification(1, "Deamidated")
        self.assertAlmostEqual(p.mass, ref_mass, 5)

    def test_picklability(self):
        for s in [p3, p8]:
            original = self.parse_sequence(s)
            duplicate = pickle.loads(pickle.dumps(original))
            self.assertEqual(
                original.total_composition(),
                duplicate.total_composition())
            self.assertEqual(original, duplicate)
            self.assertEqual(original.total_mass, duplicate.total_mass)
            self.assertEqual(original.total_composition(),
                             duplicate.total_composition())
            self.assertTrue(original.full_structure_equality(duplicate))

    def test_total_mass(self):
        seq = self.parse_sequence(p3)
        self.assertAlmostEqual(seq.total_mass, 3000.1123374719496, 3)

    def test_n_glycan_sequons(self):
        seq = self.parse_sequence(p2).deglycosylate()
        self.assertTrue(4 in seq.n_glycan_sequon_sites)

    def test_o_glycopeptide(self):
        masses = [
            718.36096708622,
            921.44033960573,
            1067.49824840467,
            1083.49316302423,
            1229.5510718231699
        ]
        for i, stub in enumerate(self.parse_sequence(p4).stub_fragments()):
            self.assertAlmostEqual(masses[i], stub.mass, 3)

    def test_sequon_finding(self):
        seq = self.parse_sequence("PEPTIDES")
        self.assertIn(3, seq.o_glycan_sequon_sites)
        self.assertNotIn(3, seq.glycosaminoglycan_sequon_sites)

    def test_composition(self):
        for p in [p1, p2, p3, p4, p5]:
            seq = self.parse_sequence(p)
            self.assertAlmostEqual(
                seq.total_mass, seq.total_composition().mass, 4)

    def test_repr_(self):
        seq = self.parse_sequence(p5)
        self.assertTrue("(Carbamidomethyl)" in (repr(seq).split('-')[0]))

    def test_glycan_composition(self):
        seq = self.parse_sequence(p3)
        self.assertEqual(seq.glycan_composition,
                         GlycanComposition(Hex=5, HexNAc=4, NeuAc=2))

    def test_delete(self):
        seq = self.parse_sequence(p1)
        ref = self.parse_sequence(p1)
        seq.delete(0)
        proline = R("P")
        self.assertAlmostEqual(
            seq.total_mass, ref.total_mass - proline.mass, 4)

    def test_fucosylated_stubs(self):
        seq = self.parse_sequence(p6)

        had_fucose = False
        for stub in seq.stub_fragments(extended=1):
            if "Fuc" in stub.name:
                had_fucose = True
        self.assertTrue(had_fucose)

    def test_multiply_glycosylated_stubs(self):
        seq = self.parse_sequence(
            "N(N-Glycosylation)ITEIVYLTN(N-Glycosylation)TTIEK{Hex:14; HexNAc:4}")
        stubs = list(seq.stub_fragments(extended=True))
        name_set = {f.name: f for f in stubs}
        self.assertTrue("peptide+Hex4HexNAc2" in name_set)
        f = name_set['peptide+Hex4HexNAc2']
        assert f.is_glycosylated
        assert f.glycosylation_size == 6
        self.assertAlmostEqual(f.glycosylation.mass(), 1072.3806, 2)

    def test_gag_linker_oxonium_ions(self):
        seq = self.parse_sequence(p7)
        self.assertAlmostEqual(seq.total_mass, 2285.766, 2)
        ox_map = {f.name: f for f in seq.glycan_fragments()}
        self.assertAlmostEqual(ox_map["HexNAca,en-Hex"].mass, 361.10089, 3)

    def test_gag_linker_stubs(self):
        seq = self.parse_sequence(p7)
        self.assertAlmostEqual(seq.total_mass, 2285.766, 2)
        peptide_mass = seq.peptide_composition().mass
        stub_map = {f.name: f for f in seq.stub_fragments()}
        bare_peptide = stub_map['peptide']
        self.assertAlmostEqual(bare_peptide.mass, peptide_mass, 2)

    def test_cad_fragmentation(self):
        t1 = self.parse_sequence(p10)
        low_energy_frags = list(t1.glycan_fragments(
            oxonium=False, all_series=True))
        assert len(low_energy_frags) == 431

    def test_glycan_representations(self):
        t1 = self.parse_sequence(p9)
        t2 = self.parse_sequence(p10)
        t3 = self.parse_sequence(p11)
        self.assertAlmostEqual(t1.total_mass, t2.total_mass, 4)
        self.assertAlmostEqual(t1.total_mass, t3.total_mass, 4)
        self.assertEqual(t1.clone().deglycosylate(),
                         t2.clone().deglycosylate())
        self.assertEqual(t1.clone().deglycosylate(),
                         t3.clone().deglycosylate())

    def test_glycan_tracking(self):
        t1 = self.parse_sequence("PEPT(HexNAc)IDE")
        t2 = self.parse_sequence("PEPT(N-Glycosylation)IDE{HexNAc:1}")
        self.assertAlmostEqual(t1.total_mass, t2.total_mass, 3)
        self.assertNotAlmostEqual(
            t1.peptide_composition().mass, t2.peptide_composition().mass, 3)

    def test_equalities(self):
        t1 = self.parse_sequence("PEPT(HexNAc)IDE")
        t2 = self.parse_sequence("PEPT(N-Glycosylation)IDE{HexNAc:1}")
        self.assertTrue(t1.base_sequence_equality(t2))
        self.assertFalse(t1.modified_sequence_equality(t2))

    def test_terminal_modification(self):
        base = self.parse_sequence("QDQC(Carbamidomethyl)IYNTTYLNVQR")
        self.assertAlmostEqual(base.mass, 1914.8894, 3)
        modified = base.clone()
        mt = modification.ModificationTable()
        pyro_glu_from_q = mt['Pyro-glu from Q']
        sites = pyro_glu_from_q.find_valid_sites(modified)
        modified.add_modification(sites[0], pyro_glu_from_q())
        self.assertAlmostEqual(modified.mass, 1897.86282353323, 3)

        base = self.parse_sequence("QDQC(Carbamidomethyl)IYNTTYLNVQR")
        p1 = base.clone()
        p1.add_modification(0, modification.Modification("TMT6plex"))
        p2 = base.clone()
        p2.n_term = base.n_term.modify(modification.Modification("TMT6plex"))
        self.assertAlmostEqual(p1.total_mass - p2.total_mass, 0)


class TestPeptideSequence(PeptideSequenceSuiteBase, unittest.TestCase):
    def parse_sequence(self, seqstr):
        return sequence.PeptideSequence(seqstr)


class TestNamedSequence(PeptideSequenceSuiteBase, unittest.TestCase):
    def parse_sequence(self, seqstr):
        return sequence.NamedSequence("spam", seqstr)


class TestAnnotatedSequence(PeptideSequenceSuiteBase, unittest.TestCase):
    def parse_sequence(self, seqstr):
        return sequence.AnnotatedSequence("spam", seqstr, annotations={"color": "green"})


if __name__ == '__main__':
    unittest.main()



from cpython cimport Py_INCREF, PY_MAJOR_VERSION
from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM, PyList_Append, PyList_GetItem, PyList_SetItem, PyList_New, PyList_SET_ITEM
from cpython.dict cimport PyDict_SetItem, PyDict_Keys, PyDict_Values
from cpython.sequence cimport PySequence_GetSlice

from glycopeptidepy._c.structure.constants cimport Configuration

from glypy.structure.glycan_composition import FrozenGlycanComposition

cdef object glycan_parser = FrozenGlycanComposition.parse

from glycopeptidepy.structure import constants as _structure_constants
from glycopeptidepy._c.compat cimport PyStr_AsString, PyStr_FromStringAndSize

cdef Configuration structure_constants = _structure_constants


# this function should use some form of linked list for accumulating positions
# when the sequence is especially long. The repeated reallocation of the list
# storing most of the sequence repeatedly seems like a likely performance bottleneck
cdef object _cstring_sequence_tokenizer(str sequence, object implicit_n_term=None, object implicit_c_term=None, object glycan_parser_function=None):
    '''A simple stateful sequence parser implementing a formally context-free language
    describing components of a polypeptide sequence with N-, C- and internal modifications,
    as well as a glycan composition written at the end.

    Return
    ------
    chunks: list
    mods: list
    n_term: str
    c_term: str
    glycan: :class:`glypy.composition.glycan_composition.FrozenGlycanComposition`
    '''
    cdef:
        ParserState state
        str current_aa, current_mod
        char next_char
        char* csequence
        object glycan
        list chunks, current_mods
        int paren_level
        size_t i, n
        Py_ssize_t mod_start, mod_end


    if glycan_parser_function is None:
        glycan_parser_function = glycan_parser

    state = ParserState.start  # [start, n_term, aa, mod, c_term]
    if implicit_n_term is None:
        n_term = structure_constants.N_TERM_DEFAULT
    else:
        n_term = implicit_n_term
    if implicit_c_term is None:
        c_term = structure_constants.C_TERM_DEFAULT
    else:
        c_term = implicit_c_term
    chunks = []
    glycan = ""
    current_aa = ""
    current_mod = ""
    current_mods = None
    paren_level = 0
    i = 0
    mod_start = -1
    mod_end = -1
    n = len(sequence)
    csequence = PyStr_AsString(sequence)
    while i < n:
        next_char = csequence[i]

        # Transition from aa to mod when encountering the start of a modification
        # internal to the sequence
        if next_char == "(":
            if state == ParserState.aa:
                state = ParserState.mod
                paren_level += 1
                mod_start = i + 1


            # Transition to n_term when starting on an open parenthesis
            elif state == ParserState.start:
                state = ParserState.n_term
                paren_level += 1
                mod_start = i + 1

            else:
                paren_level += 1
                if mod_start == -1:
                    mod_start = i + 1

        elif next_char == ")":
            if state == ParserState.aa:
                raise Exception(
                    "Invalid Sequence. ) found outside of modification, Position {0}. {1}".format(i, sequence))
            else:
                paren_level -= 1
                if paren_level == 0:
                    mod_end = i
                    current_mod = PyStr_FromStringAndSize(&csequence[mod_start], mod_end - mod_start)
                    mod_start = -1
                    mod_end = -1
                    if current_mods is None:
                        current_mods = [current_mod]
                    else:
                        current_mods.append(current_mod)
                    if state == ParserState.mod:
                        state = ParserState.aa
                        # If we encounter multiple modifications in separate parentheses
                        # in a row, attach them to the previous position
                        if current_aa == "":
                            chunks[-1][1].extend(current_mods)
                        else:
                            chunks.append([current_aa, current_mods])

                    elif state == ParserState.n_term:
                        if csequence[i + 1] != "-":
                            raise Exception("Malformed N-terminus for " + sequence)
                        # Only one modification on termini
                        n_term = current_mod
                        state = ParserState.aa
                        # Jump ahead past - into the amino acid sequence
                        i += 1

                    elif state == ParserState.c_term:
                        # Only one modification on termini
                        c_term = current_mod

                    current_mods = None
                    current_mod = ""
                    current_aa = ""
                    mod_start = -1
                    mod_end = -1
                else:
                    mod_end = i

        elif next_char == "|":
            if state == ParserState.aa:
                raise Exception("Invalid Sequence. | found outside of modification")
            else:
                mod_end = i
                current_mod = PyStr_FromStringAndSize(&csequence[mod_start], mod_end - mod_start)
                if current_mods is None:
                    current_mods = [current_mod]
                else:
                    current_mods.append(current_mod)
                mod_start = i + 1
                mod_end = -1
                current_mod = ""

        elif next_char == "{":
            if (state == ParserState.aa or (state == ParserState.c_term and paren_level == 0)):
                glycan = PyStr_FromStringAndSize(&csequence[i], n - i)
                break
            elif (state in {ParserState.mod, ParserState.n_term, ParserState.c_term}) and paren_level > 0:
                mod_end = i

        elif next_char == "-":
            if state == ParserState.aa:
                state = ParserState.c_term
                if(current_aa != ""):
                    chunks.append([current_aa, current_mods])
                    current_mod = ""
                    current_mods = None
                    current_aa = ""
            else:
                pass

        elif state == ParserState.start:
            state = ParserState.aa
            current_aa = PyStr_FromStringAndSize(&next_char, 1)
        elif state == ParserState.aa:
            if(current_aa != ""):
                chunks.append([current_aa, current_mods])
                current_mod = ""
                current_aa = ""
                current_mods = None
            current_aa = PyStr_FromStringAndSize(&next_char, 1)
        elif state in {ParserState.n_term, ParserState.mod, ParserState.c_term}:
            mod_end = i
        else:
            raise Exception(
                "Unknown Tokenizer State", current_aa, current_mod, i, next_char)
        i += 1
    if current_aa != "":
        chunks.append([current_aa, current_mods])

    if glycan != "":
        try:
            glycan = glycan_parser_function(glycan)
        except Exception as e:
            print(e, glycan)
    else:
        glycan = None

    return chunks, None, glycan, n_term, c_term

cdef object _sequence_tokenizer(sequence_encoded_t sequence, object implicit_n_term=None, object implicit_c_term=None, object glycan_parser_function=None):
    '''A simple stateful sequence parser implementing a formally context-free language
    describing components of a polypeptide sequence with N-, C- and internal modifications,
    as well as a glycan composition written at the end.

    Return
    ------
    chunks: list
    mods: list
    n_term: str
    c_term: str
    glycan: :class:`glypy.composition.glycan_composition.FrozenGlycanComposition`
    '''
    cdef:
        ParserState state
        sequence_encoded_t current_aa, current_mod, next_char
        object glycan
        list chunks, current_mods
        int paren_level, i, n

    if glycan_parser_function is None:
        glycan_parser_function = glycan_parser

    if sequence_encoded_t is str:
        return _cstring_sequence_tokenizer(sequence, implicit_n_term, implicit_c_term, glycan_parser_function)

    state = ParserState.start  # [start, n_term, aa, mod, c_term]
    if implicit_n_term is None:
        n_term = structure_constants.N_TERM_DEFAULT
    else:
        n_term = implicit_n_term
    if implicit_c_term is None:
        c_term = structure_constants.C_TERM_DEFAULT
    else:
        c_term = implicit_c_term
    chunks = []
    glycan = ""
    current_aa = ""
    current_mod = ""
    current_mods = None
    paren_level = 0
    i = 0
    n = len(sequence)
    while i < n:
        next_char = sequence[i]

        # Transition from aa to mod when encountering the start of a modification
        # internal to the sequence
        if next_char == "(":
            if state == ParserState.aa:
                state = ParserState.mod
                assert paren_level == 0
                paren_level += 1

            # Transition to n_term when starting on an open parenthesis
            elif state == ParserState.start:
                state = ParserState.n_term
                paren_level += 1

            else:
                paren_level += 1
                if not (state in {ParserState.n_term, ParserState.c_term} and paren_level == 1):
                    current_mod += next_char

        elif next_char == ")":
            if state == ParserState.aa:
                raise Exception(
                    "Invalid Sequence. ) found outside of modification, Position {0}. {1}".format(i, sequence))
            else:
                paren_level -= 1
                if paren_level == 0:
                    if current_mods is None:
                        current_mods = [current_mod]
                    else:
                        current_mods.append(current_mod)
                    if state == ParserState.mod:
                        state = ParserState.aa
                        # If we encounter multiple modifications in separate parentheses
                        # in a row, attach them to the previous position
                        if current_aa == "":
                            chunks[-1][1].extend(current_mods)
                        else:
                            chunks.append([current_aa, current_mods])

                    elif state == ParserState.n_term:
                        if sequence[i + 1] != "-":
                            raise Exception("Malformed N-terminus for " + sequence)
                        # Only one modification on termini
                        n_term = current_mod
                        state = ParserState.aa
                        # Jump ahead past - into the amino acid sequence
                        i += 1

                    elif state == ParserState.c_term:
                        # Only one modification on termini
                        c_term = current_mod

                    current_mods = None
                    current_mod = ""
                    current_aa = ""
                else:
                    current_mod += next_char

        elif next_char == "|":
            if state == ParserState.aa:
                raise Exception("Invalid Sequence. | found outside of modification")
            else:
                if current_mods is None:
                    current_mods = [current_mod]
                else:
                    current_mods.append(current_mod)
                current_mod = ""

        elif next_char == "{":
            if (state == ParserState.aa or (state == ParserState.c_term and paren_level == 0)):
                glycan = sequence[i:]
                break
            elif (state in {ParserState.mod, ParserState.n_term, ParserState.c_term}) and paren_level > 0:
                current_mod += next_char

        elif next_char == "-":
            if state == ParserState.aa:
                state = ParserState.c_term
                if(current_aa != ""):
                    if current_mods is None:
                        current_mods = [current_mod]
                    else:
                        current_mods.append(current_mod)
                    chunks.append([current_aa, current_mods])
                    current_mod = ""
                    current_mods = None
                    current_aa = ""
            else:
                current_mod += next_char

        elif state == ParserState.start:
            state = ParserState.aa
            current_aa = next_char
        elif state == ParserState.aa:
            if(current_aa != ""):
                if current_mods is None:
                    current_mods = [current_mod]
                else:
                    current_mods.append(current_mod)
                chunks.append([current_aa, current_mods])
                current_mod = ""
                current_mods = None
                current_aa = ""
            current_aa = next_char
        elif state in {ParserState.n_term, ParserState.mod, ParserState.c_term}:
            current_mod += next_char
        else:
            raise Exception(
                "Unknown Tokenizer State", current_aa, current_mod, i, next_char)
        i += 1
    if current_aa != "":
        chunks.append([current_aa, current_mods])

    if glycan != "":
        try:
            glycan = glycan_parser_function(glycan)
        except Exception as e:
            print(e, glycan)
    else:
        glycan = None

    return chunks, None, glycan, n_term, c_term


cpdef sequence_tokenizer(object sequence, object implicit_n_term=None, object implicit_c_term=None, object glycan_parser_function=None):
    if isinstance(sequence, str):
        return _sequence_tokenizer[str](<str>sequence, implicit_n_term, implicit_c_term, glycan_parser_function)
    if PY_MAJOR_VERSION > 2:
        if isinstance(sequence, bytes):
            return _sequence_tokenizer[str](
                (<bytes>sequence).decode('utf8'), implicit_n_term, implicit_c_term, glycan_parser_function)
    else:
        if isinstance(sequence, unicode):
            return _sequence_tokenizer[str](
                (<unicode>sequence).encode('utf8'), implicit_n_term, implicit_c_term, glycan_parser_function)
    return _sequence_tokenizer[object](sequence, implicit_n_term, implicit_c_term, glycan_parser_function)


cpdef parse_simple(str sequence):
    cdef:
        size_t i, n
        list chunks, chunk, junk
        str current_aa
        char* csequence

    n = len(sequence)
    chunks = PyList_New(n)
    junk = []
    csequence = PyStr_AsString(sequence)
    for i in range(n):
        next_char = csequence[i]
        if next_char == '(' or next_char == '{':
            raise ValueError("Cannot parse a modified peptide sequence with parse_simple!")
        current_aa = PyStr_FromStringAndSize(&next_char, 1)
        chunk = [current_aa, junk]
        Py_INCREF(chunk)
        PyList_SET_ITEM(chunks, i, chunk)

    return chunks, [], None, structure_constants.N_TERM_DEFAULT, structure_constants.C_TERM_DEFAULT



from typing import List
from six import string_types as basestring

from .. import constants as structure_constants
from ..base import PeptideSequenceBase, SequencePosition
from ..composition import Composition
from ..glycan import GlycosylationManager
from ..modification import Modification, ModificationCategory
from ..residue import Residue
from ..terminal_group import TerminalGroup, _make_terminal_group
from ..parser import sequence_tokenizer


def _total_composition(sequence):
    total = Composition()
    for position in sequence:
        total += position[0].composition
        for mod in position[1]:
            if mod.is_tracked_for(ModificationCategory.glycosylation):
                continue
            total += mod.composition
    total += sequence.n_term.composition
    total += sequence.c_term.composition
    gc = sequence.glycan_composition
    if gc:
        total += gc.total_composition()
    return total


class _PeptideSequenceCore(PeptideSequenceBase):
    sequence: List[SequencePosition]
    mass: float
    _glycosylation_manager: GlycosylationManager

    _n_term: TerminalGroup
    _c_term: TerminalGroup


    def __init__(self, sequence=None, parser_function=None, **kwargs):
        if parser_function is None:
            parser_function = sequence_tokenizer
        self._initialize_fields()
        self._init_from_string(sequence, parser_function, **kwargs)

    def _init_from_parsed_string(self, seq_list, glycan=None, n_term=None, c_term=None, **kwargs):
        i = 0
        self.sequence = [None for _ in seq_list]
        mass = 0
        for item in seq_list:
            res = Residue.parse(item[0])
            mass += res.mass
            mods = []
            for mod in item[1]:
                if mod != '':
                    mod = Modification(mod)
                    if mod.is_tracked_for(ModificationCategory.glycosylation):
                        self._glycosylation_manager[i] = mod
                    mods.append(mod)
                    mass += mod.mass
            self.sequence[i] = SequencePosition([res, mods])
            i += 1
        self.mass = mass
        has_glycan = glycan is not None
        if has_glycan:
            self._glycosylation_manager.aggregate = glycan

        if isinstance(n_term, basestring):
            if n_term != structure_constants.N_TERM_DEFAULT:
                n_term = Modification(n_term)
            else:
                n_term = None
        if isinstance(c_term, basestring):
            if c_term != structure_constants.C_TERM_DEFAULT:
                c_term = Modification(c_term)
            else:
                c_term = None
        n_term_group = _make_terminal_group(structure_constants.N_TERM_DEFAULT, n_term)
        c_term_group = _make_terminal_group(structure_constants.C_TERM_DEFAULT, c_term)
        self._init_termini(n_term_group, c_term_group)

    def _init_from_string(self, sequence, parser_function, **kwargs):
        """Initialize a :class:`PeptideSequence` from a parse-able string.

        Parameters
        ----------
        sequence : :class:`str`
            The string to parse
        parser_function : :class:`Callable`
            The parsing function to use
        """
        if sequence == "" or sequence is None:
            pass
        else:
            seq_list, modifications, glycan, n_term, c_term = parser_function(
                sequence)
            self._init_from_parsed_string(seq_list, glycan, n_term, c_term, **kwargs)

    def _init_from_components(self, seq_list, glycan_composition=None, n_term=None, c_term=None, **kwargs):
        """Initialize a :class:`PeptideSequence`'s state from a sequence of (
        :class:`~.AminoAcidResidue`, [:class:`~.Modification`, ]) pairs, with optional extra information
        defining the glycan composition aggregate and the terminal groups

        Parameters
        ----------
        seq_list : :class:`Sequence`
            The sequence of (:class:`~.AminoAcidResidue`, [:class:`~.Modification`, ])
        glycan_composition : :class:`~.GlycanComposition`, optional
            The aggregate glycan composition associated with the sequence not fully specified
            by the modifications in ``seq_list``.
        n_term : :class:`~.Modification`, optional
            The N-terminal modification, if any
        c_term : :class:`~.Modification`, optional
            The C-terminal modification, if any
        """
        mass = 0
        self.sequence = [None for _ in seq_list]
        i = 0
        for res, mods in seq_list:
            mass += res.mass
            for mod in mods:
                if mod.is_tracked_for(ModificationCategory.glycosylation):
                    self._glycosylation_manager[i] = mod
                mass += mod.mass
            self.sequence[i] = SequencePosition([res, list(mods)])
            i += 1
        self.mass = mass
        if glycan_composition is not None:
            self._glycosylation_manager.aggregate = glycan_composition.clone()
        self._init_termini(
            _make_terminal_group(structure_constants.N_TERM_DEFAULT, n_term),
            _make_terminal_group(structure_constants.C_TERM_DEFAULT, c_term))

    def _init_termini(self, n_term, c_term):
        self._n_term = n_term
        self._c_term = c_term
        self._mass += n_term.mass + c_term.mass

    def _initialize_fields(self):
        """Initialize all mutable fields to their default, empty values.
        """
        self._mass = 0.0
        self.sequence = []
        self._fragment_index = None

        self._glycosylation_manager = GlycosylationManager(self)

        self._n_term = None
        self._c_term = None

        self._total_composition = None
        self._peptide_composition = None

    def _invalidate(self):
        self._total_composition = None
        self._peptide_composition = None

    def __repr__(self):
        n_term = ""
        if self.n_term is not None:
            n_term = "({0})-".format(self.n_term)
        c_term = ""
        if self.c_term is not None:
            c_term = "-({0})".format(self.c_term)
        aggregate = self._glycosylation_manager.aggregate if self._glycosylation_manager.aggregate is not None else ""
        rep = "{n_term}{sequence}{c_term}{glycan_aggregate}[{_mass}]".format(
            n_term=n_term, c_term=c_term,
            glycan_aggregate=aggregate,
            sequence=self.sequence,
            _mass=self._mass)
        return rep

    @property
    def n_term(self):
        return self._n_term

    @n_term.setter
    def n_term(self, value):
        if isinstance(value, Modification):
            value = self._n_term.modify(value)
        self._invalidate()
        reset_mass = 0
        try:
            reset_mass = self._n_term.mass
        except AttributeError:
            pass
        self._n_term = value
        new_mass = 0
        try:
            new_mass = value.mass
        except AttributeError:
            pass
        self.mass += new_mass - reset_mass

    @property
    def c_term(self):
        return self._c_term

    @c_term.setter
    def c_term(self, value):
        if isinstance(value, Modification):
            value = self._c_term.modify(value)
        self._invalidate()
        reset_mass = 0
        try:
            reset_mass = self._c_term.mass
        except AttributeError:
            pass
        self._c_term = value
        new_mass = 0
        try:
            new_mass = value.mass
        except AttributeError:
            pass
        self.mass += new_mass - reset_mass

    @property
    def mass(self):
        return self._mass

    @mass.setter
    def mass(self, value):
        self._mass = value

    @property
    def peptide_backbone_mass(self) -> float:
        return self.peptide_composition().mass

    @property
    def total_mass(self) -> float:
        total = 0
        for position in self:
            total += position[0].mass
            for mod in position[1]:
                if mod.is_tracked_for(ModificationCategory.glycosylation):
                    continue
                total += mod.mass
        total += self.n_term.mass
        total += self.c_term.mass
        gc = self.glycan_composition
        if gc:
            total += gc.mass()
        return total

    def peptide_composition(self) -> Composition:
        if self._peptide_composition is None:
            if self.glycan is None:
                glycan_composition = Composition()
            else:
                glycan_composition = self.glycan_composition.total_composition()
            self._peptide_base_composition = self.total_composition() - glycan_composition
        return self._peptide_base_composition

    def total_composition(self) -> Composition:
        if self._total_composition is None:
            self._total_composition = _total_composition(self)
        return self._total_composition

    def __iter__(self):
        return iter(self.sequence)

    def __getitem__(self, index):
        sub = self.sequence[index]
        return sub

    def __setitem__(self, index, value):
        self._invalidate()
        self.sequence[index] = value

    def __len__(self):
        return len(self.sequence)

    # equality methods

    def __eq__(self, other):
        return str(self) == str(other)

    def base_sequence_equality(self, other):
        if len(self) != len(other):
            return False
        for a, b in zip(self, other):
            if a[0] != b[0]:
                return False
        return True

    def modified_sequence_equality(self, other):
        if len(self) != len(other):
            return False
        for a, b in zip(self, other):
            if a[0] != b[0] or set(a[1]) != set(b[1]):
                return False
        return True

    def full_structure_equality(self, other):
        sequences_equal = self.modified_sequence_equality(other)
        if sequences_equal:
            return self.glycan == other.glycan
        else:
            return False

    def __ne__(self, other):
        return str(self) != str(other)

    def __hash__(self):
        return hash(str(self))

    def get_sequence(self, include_glycan=True, include_termini=True, implicit_n_term=None, implicit_c_term=None):
        """
        Generate human readable sequence string. Called by :meth:`__str__`

        Parameters
        ----------
        include_glycan: bool
            Whether to include the glycan in the resulting string. Defaults to `True`
        include_termini: bool
            Whether to include the N- and C-termini. Make sure this is `True` if you want non-standard
            termini to be properly propagated.


        Returns
        -------
        str
        """
        if implicit_n_term is None:
            implicit_n_term = structure_constants.N_TERM_DEFAULT
        if implicit_c_term is None:
            implicit_c_term = structure_constants.C_TERM_DEFAULT

        seq_list = []
        for x, y in self.sequence:
            mod_str = ''
            if y:
                mod_str = '|'.join(mod.serialize() for mod in y)
                mod_str = ''.join(['(', mod_str, ')'])
            seq_list.append(''.join([x.symbol, mod_str]))
        rep = ''.join(seq_list)
        if include_termini:
            n_term = ""
            if self.n_term is not None and self.n_term != implicit_n_term:
                n_term = "({0})-".format(self.n_term.serialize())
            c_term = ""
            if self.c_term is not None and self.c_term != implicit_c_term:
                c_term = "-({0})".format(self.c_term.serialize())
            rep = "{0}{1}{2}".format(n_term, rep, c_term)
        if include_glycan:
            if self._glycosylation_manager.aggregate is not None:
                rep += str(self._glycosylation_manager.aggregate)
        return rep

    def __str__(self):
        return self.get_sequence()

    def clone(self):
        inst = self.__class__()
        if self._glycosylation_manager.aggregate is not None:
            glycan = self._glycosylation_manager.aggregate.clone()
            glycan.composition_offset = Composition("H2O")
        else:
            glycan = None
        inst._init_from_components(
            self.sequence, glycan,
            self.n_term.modification,
            self.c_term.modification)
        return inst


try:
    from glycopeptidepy._c.structure import sequence_methods
    _PeptideSequenceCore = sequence_methods._PeptideSequenceCore
except ImportError as err:
    print(err)
